#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants used by all sub-modules of dpxdt.

* Do not import anything into this file!
* Do not put configuration in this file!
* Do not put secret keys in this file!
"""


CAPTURE_QUEUE_NAME = 'capture'

PDIFF_QUEUE_NAME = 'run-pdiff'

SITE_DIFF_QUEUE_NAME = 'site-diff'



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Workers for driving screen captures, perceptual diffs, and related work."""

import Queue
import logging
import subprocess
import time

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

# Local modules
from dpxdt.client import timer_worker
from dpxdt.client import workers


class Error(Exception):
    """Base class for exceptions in this module."""

class TimeoutError(Exception):
    """Subprocess has taken too long to complete and was terminated."""


class ProcessWorkflow(workers.WorkflowItem):
    """Workflow that runs a subprocess.

    Args:
        log_path: Path to where output from this subprocess should be written.
        timeout_seconds: How long before the process should be force killed.

    Returns:
        The return code of the subprocess.
    """

    def get_args(self):
        """Return the arguments for running the subprocess."""
        raise NotImplemented

    def run(self, log_path, timeout_seconds=30):
        start_time = time.time()
        with open(log_path, 'a') as output_file:
            args = self.get_args()
            logging.info('item=%r Running subprocess: %r', self, args)
            try:
                process = subprocess.Popen(
                    args,
                    stderr=subprocess.STDOUT,
                    stdout=output_file,
                    close_fds=True)
            except:
                logging.error('item=%r Failed to run subprocess: %r',
                              self, args)
                raise

            while True:
                logging.info('item=%r Polling pid=%r', self, process.pid)
                # NOTE: Use undocumented polling method to work around a
                # bug in subprocess for handling defunct zombie processes:
                # http://bugs.python.org/issue2475
                process._internal_poll(_deadstate=127)
                if process.returncode is not None:
                    logging.info(
                        'item=%r Subprocess finished pid=%r, returncode=%r',
                        self, process.pid, process.returncode)
                    raise workers.Return(process.returncode)

                now = time.time()
                run_time = now - start_time
                if run_time > timeout_seconds:
                    logging.info('item=%r Subprocess timed out pid=%r',
                                 self, process.pid)
                    process.kill()
                    raise TimeoutError(
                        'Sent SIGKILL to item=%r, pid=%s, run_time=%s' %
                        (self, process.pid, run_time))

                yield timer_worker.TimerItem(FLAGS.polltime)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility for doing incremental diffs for a live website.

Example usage:

./dpxdt/tools/site_diff.py \
    --upload_build_id=1234 \
    --release_server_prefix=https://my-dpxdt-apiserver.example.com/api \
    --release_client_id=<your api key> \
    --release_client_secret=<your api secret> \
    --crawl_depth=1 \
    http://www.example.com/my/website/here
"""

import HTMLParser
import Queue
import datetime
import json
import logging
import os
import re
import sys
import urlparse

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

# Local modules
from dpxdt.client import release_worker
from dpxdt.client import fetch_worker
from dpxdt.client import workers
import flags


gflags.DEFINE_integer(
    'crawl_depth', -1,
    'How deep to crawl. Depth of 0 means only the given page. 1 means pages '
    'that are one click away, 2 means two clicks, and so on. Set to -1 to '
    'scan every URL with the supplied prefix.')

gflags.DEFINE_spaceseplist(
    'ignore_prefixes', [],
    'URL prefixes that should not be crawled.')


# URL regex rewriting code originally from mirrorrr
# http://code.google.com/p/mirrorrr/source/browse/trunk/transform_content.py

# URLs that have absolute addresses
ABSOLUTE_URL_REGEX = r"(?P<url>(http(s?):)?//[^\"'> \t]+)"
# URLs that are relative to the base of the current hostname.
BASE_RELATIVE_URL_REGEX = (
    r"/(?!(/)|(http(s?)://)|(url\())(?P<url>[^\"'> \t]*)")
# URLs that have '../' or './' to start off their paths.
TRAVERSAL_URL_REGEX = (
    r"(?P<relative>\.(\.)?)/(?!(/)|"
    r"(http(s?)://)|(url\())(?P<url>[^\"'> \t]*)")
# URLs that are in the same directory as the requested URL.
SAME_DIR_URL_REGEX = (
    r"(?!(/)|([a-z0-9.-]{2,10}:(//)?)|(#)|(url\())(?P<url>[^\\\"'> \t]+)")
# URL matches the root directory.
ROOT_DIR_URL_REGEX = r"(?!//(?!>))/(?P<url>)(?=[ \t\n]*[\"'> /])"
# Start of a tag using 'src' or 'href'
TAG_START = (
    r"(?i)(?P<tag><?[^\s]*)(?P<attr>\s(src|href|action|url|background))"
    r"(?P<equals>[\t ]*=[\t ]*)(?P<quote>[\"']?)")
# Potential HTML document URL with no fragments.
MAYBE_HTML_URL_REGEX = (
    TAG_START + r"(?P<absurl>(http(s?):)?//[^\"'> \t]+)")

REPLACEMENT_REGEXES = [
    (TAG_START + SAME_DIR_URL_REGEX,
     "\g<tag>\g<attr>\g<equals>\g<quote>%(accessed_dir)s\g<url>"),
    (TAG_START + TRAVERSAL_URL_REGEX,
     "\g<tag>\g<attr>\g<equals>\g<quote>%(accessed_dir)s/\g<relative>/\g<url>"),
    (TAG_START + BASE_RELATIVE_URL_REGEX,
     "\g<tag>\g<attr>\g<equals>\g<quote>%(base)s/\g<url>"),
    (TAG_START + ROOT_DIR_URL_REGEX,
     "\g<tag>\g<attr>\g<equals>\g<quote>%(base)s/"),
    (TAG_START + ABSOLUTE_URL_REGEX,
     "\g<tag>\g<attr>\g<equals>\g<quote>\g<url>"),
]


def clean_url(url, force_scheme=None):
    """Cleans the given URL."""
    # URL should be ASCII according to RFC 3986
    url = str(url)
    # Collapse ../../ and related
    url_parts = urlparse.urlparse(url)
    path_parts = []
    for part in url_parts.path.split('/'):
        if part == '.':
            continue
        elif part == '..':
            if path_parts:
                path_parts.pop()
        else:
            path_parts.append(part)

    url_parts = list(url_parts)
    if force_scheme:
        url_parts[0] = force_scheme
    url_parts[2] = '/'.join(path_parts)
    url_parts[4] = ''    # No query string
    url_parts[5] = ''    # No path

    # Always have a trailing slash
    if not url_parts[2]:
        url_parts[2] = '/'

    return urlparse.urlunparse(url_parts)


def extract_urls(url, data, unescape=HTMLParser.HTMLParser().unescape):
    """Extracts the URLs from an HTML document."""
    parts = urlparse.urlparse(url)
    prefix = '%s://%s' % (parts.scheme, parts.netloc)

    accessed_dir = os.path.dirname(parts.path)
    if not accessed_dir.endswith('/'):
        accessed_dir += '/'

    for pattern, replacement in REPLACEMENT_REGEXES:
        fixed = replacement % {
            'base': prefix,
            'accessed_dir': accessed_dir,
        }
        data = re.sub(pattern, fixed, data)

    result = set()
    for match in re.finditer(MAYBE_HTML_URL_REGEX, data):
        link = match.groupdict()
        # Ignore JavaScript variable assignments
        if 'var' == link['tag'] and '' == link['quote']:
            continue
        found_url = unescape(link['absurl'])
        found_url = clean_url(
            found_url,
            force_scheme=parts[0])  # Use the main page's scheme
        result.add(found_url)

    return result


IGNORE_SUFFIXES = frozenset([
    'jpg', 'jpeg', 'png', 'css', 'js', 'xml', 'json', 'gif', 'ico', 'doc'])


def prune_urls(url_set, start_url, allowed_list, ignored_list):
    """Prunes URLs that should be ignored."""
    result = set()

    for url in url_set:
        allowed = False
        for allow_url in allowed_list:
            if url.startswith(allow_url):
                allowed = True
                break

        if not allowed:
            continue

        ignored = False
        for ignore_url in ignored_list:
            if url.startswith(ignore_url):
                ignored = True
                break

        if ignored:
            continue

        prefix, suffix = (url.rsplit('.', 1) + [''])[:2]
        if suffix.lower() in IGNORE_SUFFIXES:
            continue

        result.add(url)

    return result


class SiteDiff(workers.WorkflowItem):
    """Workflow for coordinating the site diff.

    Args:
        start_url: URL to begin the site diff scan.
        ignore_prefixes: Optional. List of URL prefixes to ignore during
            the crawl; start_url should be a common prefix with all of these.
        upload_build_id: Build ID of the site being compared.
        upload_release_name: Optional. Release name to use for the build. When
            not supplied, a new release based on the current time will be
            created.
        heartbeat: Function to call with progress status.
    """

    def run(self,
            start_url,
            ignore_prefixes,
            upload_build_id,
            upload_release_name=None,
            heartbeat=None):
        if not ignore_prefixes:
            ignore_prefixes = []

        pending_urls = set([clean_url(start_url)])
        seen_urls = set()
        good_urls = set()

        yield heartbeat('Scanning for content')

        limit_depth = FLAGS.crawl_depth >= 0
        depth = 0
        while (not limit_depth or depth <= FLAGS.crawl_depth) and pending_urls:
            # TODO: Enforce a job-wide timeout on the whole process of
            # URL discovery, to make sure infinitely deep sites do not
            # cause this job to never stop.
            seen_urls.update(pending_urls)
            yield heartbeat(
                'Scanning %d pages for good urls' % len(pending_urls))
            output = yield [fetch_worker.FetchItem(u) for u in pending_urls]
            pending_urls.clear()

            for item in output:
                if not item.data:
                    logging.debug('No data from url=%r', item.url)
                    continue

                if item.headers.gettype() != 'text/html':
                    logging.debug('Skipping non-HTML document url=%r',
                                  item.url)
                    continue

                good_urls.add(item.url)
                found = extract_urls(item.url, item.data)
                pruned = prune_urls(
                    found, start_url, [start_url], ignore_prefixes)
                new = pruned - seen_urls
                pending_urls.update(new)
                yield heartbeat('Found %d new URLs from %s' % (
                                len(new), item.url))

            yield heartbeat('Finished crawl at depth %d' % depth)
            depth += 1

        yield heartbeat(
            'Found %d total URLs, %d good HTML pages; starting '
            'screenshots' % (len(seen_urls), len(good_urls)))

        # TODO: Make the default release name prettier.
        if not upload_release_name:
            upload_release_name = str(datetime.datetime.utcnow())

        release_number = yield release_worker.CreateReleaseWorkflow(
            upload_build_id, upload_release_name, start_url)

        run_requests = []
        for url in good_urls:
            yield heartbeat('Requesting run for %s' % url)
            parts = urlparse.urlparse(url)
            run_name = parts.path

            config_dict = {
                'viewportSize': {
                    'width': 1280,
                    'height': 1024,
                }
            }
            if FLAGS.inject_css:
                config_dict['injectCss'] = FLAGS.inject_css
            if FLAGS.inject_js:
                config_dict['injectJs'] = FLAGS.inject_js
            if FLAGS.cookies:
                config_dict['cookies'] = json.loads(
                    open(FLAGS.cookies).read())
            config_data = json.dumps(config_dict)

            run_requests.append(release_worker.RequestRunWorkflow(
                upload_build_id, upload_release_name, release_number,
                run_name, url, config_data))

        yield run_requests

        yield heartbeat('Marking runs as complete')
        release_url = yield release_worker.RunsDoneWorkflow(
            upload_build_id, upload_release_name, release_number)

        yield heartbeat('Results viewable at: %s' % release_url)


def real_main(start_url=None,
              ignore_prefixes=None,
              upload_build_id=None,
              upload_release_name=None):
    """Runs the site_diff."""
    coordinator = workers.get_coordinator()
    fetch_worker.register(coordinator)
    coordinator.start()

    item = SiteDiff(
        start_url=start_url,
        ignore_prefixes=ignore_prefixes,
        upload_build_id=upload_build_id,
        upload_release_name=upload_release_name,
        heartbeat=workers.PrintWorkflow)
    item.root = True

    coordinator.input_queue.put(item)
    coordinator.wait_one()
    coordinator.stop()
    coordinator.join()


def main(argv):
    try:
        argv = FLAGS(argv)
    except gflags.FlagsError, e:
        print '%s\nUsage: %s ARGS\n%s' % (e, sys.argv[0], FLAGS)
        sys.exit(1)

    if len(argv) != 2:
        print 'Must supply a website URL as the first argument.'
        sys.exit(1)

    assert FLAGS.upload_build_id
    assert FLAGS.release_server_prefix

    if FLAGS.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    real_main(
        start_url=argv[1],
        ignore_prefixes=FLAGS.ignore_prefixes,
        upload_build_id=FLAGS.upload_build_id,
        upload_release_name=FLAGS.upload_release_name)


if __name__ == '__main__':
    main(sys.argv)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Workers for driving screen captures, perceptual diffs, and related work."""

import Queue
import heapq
import logging
import time

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

# Local modules
from dpxdt.client import workers


class TimerItem(workers.WorkItem):
    """Work item for waiting some period of time before returning."""

    def __init__(self, delay_seconds):
        workers.WorkItem.__init__(self)
        self.delay_seconds = delay_seconds
        self.ready_time = time.time() + delay_seconds


class TimerThread(workers.WorkerThread):
    """"Worker thread that tracks many timers."""

    def __init__(self, *args):
        """Initializer."""
        workers.WorkerThread.__init__(self, *args)
        self.timers = []

    def handle_nothing(self):
        now = time.time()
        while self.timers:
            ready_time, _ = self.timers[0]
            wait_time = ready_time - now
            if wait_time <= 0:
                _, item = heapq.heappop(self.timers)
                self.output_queue.put(item)
            else:
                # Wait for new work up to the point that the earliest
                # timer is ready to fire.
                self.polltime = wait_time
                return

        # Nothing to do, use the default poll time.
        self.polltime = FLAGS.polltime

    def handle_item(self, item):
        heapq.heappush(self.timers, (item.ready_time, item))
        self.handle_nothing()


def register(coordinator):
    """Registers this module as a worker with the given coordinator."""
    timer_queue = Queue.Queue()
    coordinator.register(TimerItem, timer_queue)
    coordinator.worker_threads.append(
        TimerThread(timer_queue, coordinator.input_queue))



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cacheable operations and eviction for models in the frontend."""

import functools
import logging

# Local libraries
import sqlalchemy

# Local modules
from . import app
from . import cache
from . import db
from dpxdt.server import models
from dpxdt.server import signals
from dpxdt.server import utils
from dpxdt.server import work_queue


class UserOps(object):
    """Cacheable operations for user-specific information."""

    def __init__(self, user_id):
        self.user_id = user_id

    # For Flask-Cache keys
    def __repr__(self):
        return 'caching.UserOps(user_id=%r)' % self.user_id

    @cache.memoize(per_instance=True)
    def load(self):
        if not self.user_id:
            return None
        user = models.User.query.get(self.user_id)
        if user:
            db.session.expunge(user)
        return user

    @cache.memoize(per_instance=True)
    def get_builds(self):
        if self.user_id:
            user = models.User.query.get(self.user_id)
            build_list = (
                user.builds
                .order_by(models.Build.created.desc())
                .limit(1000)
                .all())
        else:
            # Anonymous users see only public builds
            build_list = (
                models.Build.query
                .filter_by(public=True)
                .order_by(models.Build.created.desc())
                .limit(1000)
                .all())

        for build in build_list:
            db.session.expunge(build)

        return build_list

    @cache.memoize(per_instance=True)
    def owns_build(self, build_id):
        build = models.Build.query.get(build_id)
        user_is_owner = False
        if build:
            user_is_owner = build.is_owned_by(self.user_id)
            db.session.expunge(build)

        return build, user_is_owner

    def evict(self):
        """Evict all caches related to this user."""
        logging.debug('Evicting cache for %r', self)
        cache.delete_memoized(self.load)
        cache.delete_memoized(self.get_builds)
        cache.delete_memoized(self.owns_build)


class ApiKeyOps(object):
    """Cacheable operations for API key-specific information."""

    def __init__(self, client_id, client_secret):
        self.client_id = client_id
        self.client_secret = client_secret

    # For Flask-Cache keys
    def __repr__(self):
        return 'caching.ApiKeyOps(client_id=%r)' % self.client_id

    @cache.memoize(per_instance=True)
    def get(self):
        api_key = models.ApiKey.query.get(self.client_id)
        utils.jsonify_assert(api_key, 'API key must exist', 403)
        utils.jsonify_assert(api_key.active, 'API key must be active', 403)
        utils.jsonify_assert(api_key.secret == self.client_secret,
                             'Must have good credentials', 403)
        return api_key

    @cache.memoize(per_instance=True)
    def can_access_build(self, build_id):
        api_key = self.get()

        build = models.Build.query.get(build_id)
        utils.jsonify_assert(build is not None, 'build must exist', 404)

        if not api_key.superuser:
            utils.jsonify_assert(api_key.build_id == build_id,
                                 'API key must have access', 404)

        return api_key, build

    def evict(self):
        """Evict all caches related to this API key."""
        logging.debug('Evicting cache for %r', self)
        cache.delete_memoized(self.get)
        cache.delete_memoized(self.can_access_build)


class BuildOps(object):
    """Cacheable operations for build-specific operations."""

    def __init__(self, build_id):
        self.build_id = build_id

    # For Flask-Cache keys
    def __repr__(self):
        return 'caching.BuildOps(build_id=%r)' % self.build_id

    @staticmethod
    def sort_run(run):
        """Sort function for runs within a release."""
        # Sort errors first, then by name. Also show errors that were manually
        # approved, so the paging sort order stays the same even after users
        # approve a diff on the run page.
        if run.status in models.Run.DIFF_NEEDED_STATES:
            return (0, run.name)
        return (1, run.name)

    @staticmethod
    def get_stats_keys(status):
        if status in (models.Run.DIFF_APPROVED,
                      models.Run.DIFF_NOT_FOUND):
            return ('runs_successful', 'runs_complete', 'runs_total')
        elif status in models.Run.DIFF_FOUND:
            return ('runs_failed', 'runs_complete', 'runs_total')
        elif status == models.Run.NO_DIFF_NEEDED:
            return ('runs_baseline',)
        elif status == models.Run.NEEDS_DIFF:
            return ('runs_total', 'runs_pending')
        elif status == models.Run.FAILED:
            return ('runs_failed',)
        return ('runs_pending',)

    @cache.memoize(per_instance=True)
    def get_candidates(self, page_size, offset):
        candidate_list = (
            models.Release.query
            .filter_by(build_id=self.build_id)
            .order_by(models.Release.created.desc())
            .offset(offset)
            .limit(page_size + 1)
            .all())

        stats_counts = []

        has_next_page = len(candidate_list) > page_size
        if has_next_page:
            candidate_list = candidate_list[:-1]

        if candidate_list:
            candidate_keys = [c.id for c in candidate_list]
            stats_counts = (
                db.session.query(
                    models.Run.release_id,
                    models.Run.status,
                    sqlalchemy.func.count(models.Run.id))
                .join(models.Release)
                .filter(models.Release.id.in_(candidate_keys))
                .group_by(models.Run.status, models.Run.release_id)
                .all())

        for candidate in candidate_list:
            db.session.expunge(candidate)

        return has_next_page, candidate_list, stats_counts

    @cache.memoize(per_instance=True)
    def get_release(self, release_name, release_number):
        release = (
            models.Release.query
            .filter_by(
                build_id=self.build_id,
                name=release_name,
                number=release_number)
            .first())

        if not release:
            return None, None, None, None

        run_list = list(release.runs)
        run_list.sort(key=BuildOps.sort_run)

        stats_dict = dict(
            runs_total=0,
            runs_complete=0,
            runs_successful=0,
            runs_failed=0,
            runs_baseline=0,
            runs_pending=0)
        for run in run_list:
            for key in self.get_stats_keys(run.status):
                stats_dict[key] += 1

        approval_log = None
        if release.status in (models.Release.GOOD, models.Release.BAD):
            approval_log = (
                models.AdminLog.query
                .filter_by(release_id=release.id)
                .filter(models.AdminLog.log_type.in_(
                    (models.AdminLog.RELEASE_BAD,
                     models.AdminLog.RELEASE_GOOD)))
                .order_by(models.AdminLog.created.desc())
                .first())

        for run in run_list:
            db.session.expunge(run)

        if approval_log:
            db.session.expunge(approval_log)

        return release, run_list, stats_dict, approval_log

    def _get_next_previous_runs(self, run):
        next_run = None
        previous_run = None

        # We sort the runs in the release by diffs first, then by name.
        # Simulate that behavior here with multiple queries.
        if run.status in models.Run.DIFF_NEEDED_STATES:
            previous_run = (
                models.Run.query
                .filter_by(release_id=run.release_id)
                .filter(models.Run.status.in_(models.Run.DIFF_NEEDED_STATES))
                .filter(models.Run.name < run.name)
                .order_by(models.Run.name.desc())
                .first())
            next_run = (
                models.Run.query
                .filter_by(release_id=run.release_id)
                .filter(models.Run.status.in_(models.Run.DIFF_NEEDED_STATES))
                .filter(models.Run.name > run.name)
                .order_by(models.Run.name)
                .first())

            if not next_run:
                next_run = (
                    models.Run.query
                    .filter_by(release_id=run.release_id)
                    .filter(
                        ~models.Run.status.in_(models.Run.DIFF_NEEDED_STATES))
                    .order_by(models.Run.name)
                    .first())
        else:
            previous_run = (
                models.Run.query
                .filter_by(release_id=run.release_id)
                .filter(~models.Run.status.in_(models.Run.DIFF_NEEDED_STATES))
                .filter(models.Run.name < run.name)
                .order_by(models.Run.name.desc())
                .first())
            next_run = (
                models.Run.query
                .filter_by(release_id=run.release_id)
                .filter(~models.Run.status.in_(models.Run.DIFF_NEEDED_STATES))
                .filter(models.Run.name > run.name)
                .order_by(models.Run.name)
                .first())

            if not previous_run:
                previous_run = (
                    models.Run.query
                    .filter_by(release_id=run.release_id)
                    .filter(
                        models.Run.status.in_(models.Run.DIFF_NEEDED_STATES))
                    .order_by(models.Run.name.desc())
                    .first())

        return next_run, previous_run

    @cache.memoize(per_instance=True)
    def get_run(self, release_name, release_number, test_name):
        run = (
            models.Run.query
            .join(models.Release)
            .filter(models.Release.name == release_name)
            .filter(models.Release.number == release_number)
            .filter(models.Run.name == test_name)
            .first())
        if not run:
            return None, None, None, None

        next_run, previous_run = self._get_next_previous_runs(run)

        approval_log = None
        if run.status == models.Run.DIFF_APPROVED:
            approval_log = (
                models.AdminLog.query
                .filter_by(run_id=run.id,
                           log_type=models.AdminLog.RUN_APPROVED)
                .order_by(models.AdminLog.created.desc())
                .first())

        if run:
            db.session.expunge(run)
        if next_run:
            db.session.expunge(next_run)
        if previous_run:
            db.session.expunge(previous_run)
        if approval_log:
            db.session.expunge(approval_log)

        return run, next_run, previous_run, approval_log

    def evict(self):
        """Evict all caches relating to this build."""
        logging.debug('Evicting cache for %r', self)
        cache.delete_memoized(self.get_candidates)
        cache.delete_memoized(self.get_release)
        cache.delete_memoized(self.get_run)



# Connect Frontend and API events to cache eviction.


def _evict_user_cache(sender, user=None, build=None):
    UserOps(user.get_id()).evict()


def _evict_build_cache(sender, build=None, release=None, run=None):
    BuildOps(build.id).evict()


def _evict_task_cache(sender, task=None):
    if not task.run_id:
        return
    run = models.Run.query.get(task.run_id)
    # Update the modification time on the run, since the task status changed.
    db.session.add(run)
    BuildOps(run.release.build_id).evict()


signals.build_updated.connect(_evict_user_cache, app)
signals.release_updated_via_api.connect(_evict_build_cache, app)
signals.run_updated_via_api.connect(_evict_build_cache, app)
signals.task_updated.connect(_evict_task_cache, app)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility for doing a diff between a pair of URLs.

Example usage:

./dpxdt/tools/url_pair_diff.py \
    --upload_build_id=1234 \
    --release_server_prefix=https://my-dpxdt-apiserver.example.com/api \
    --release_client_id=<your api key> \
    --release_client_secret=<your api secret> \
    http://www.example.com/my/before/page \
    http://www.example.com/my/after/page
"""

import HTMLParser
import Queue
import datetime
import json
import logging
import os
import re
import shutil
import sys
import tempfile
import urlparse

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

# Local modules
from dpxdt.client import fetch_worker
from dpxdt.client import release_worker
from dpxdt.client import workers
import flags


class UrlPairDiff(workers.WorkflowItem):
    """Workflow for diffing a pair of URLs.

    Args:
        new_url: URL of the newest and best version of the page.
        baseline_url: URL of the baseline to compare to.
        upload_build_id: Optional. Build ID of the site being compared. When
            supplied a new release will be cut for this build comparing it
            to the last good release.
        upload_release_name: Optional. Release name to use for the build. When
            not supplied, a new release based on the current time will be
            created.
        heartbeat: Function to call with progress status.
    """

    def run(self,
            new_url,
            baseline_url,
            upload_build_id,
            upload_release_name=None,
            heartbeat=None):
        # TODO: Make the default release name prettier.
        if not upload_release_name:
            upload_release_name = str(datetime.datetime.utcnow())

        yield heartbeat('Creating release %s' % upload_release_name)
        release_number = yield release_worker.CreateReleaseWorkflow(
            upload_build_id, upload_release_name, new_url)

        config_dict = {
            'viewportSize': {
                'width': 1280,
                'height': 1024,
            }
        }
        if FLAGS.inject_css:
            config_dict['injectCss'] = FLAGS.inject_css
        if FLAGS.inject_js:
            config_dict['injectJs'] = FLAGS.inject_js
        config_data = json.dumps(config_dict)

        url_parts = urlparse.urlparse(new_url)

        yield heartbeat('Requesting captures')
        yield release_worker.RequestRunWorkflow(
                upload_build_id,
                upload_release_name,
                release_number,
                url_parts.path or '/',
                new_url,
                config_data,
                ref_url=baseline_url,
                ref_config_data=config_data)

        yield heartbeat('Marking runs as complete')
        release_url = yield release_worker.RunsDoneWorkflow(
            upload_build_id, upload_release_name, release_number)

        yield heartbeat('Results viewable at: %s' % release_url)


def real_main(new_url=None,
              baseline_url=None,
              upload_build_id=None,
              upload_release_name=None):
    """Runs the ur_pair_diff."""
    coordinator = workers.get_coordinator()
    fetch_worker.register(coordinator)
    coordinator.start()

    item = UrlPairDiff(
        new_url,
        baseline_url,
        upload_build_id,
        upload_release_name=upload_release_name,
        heartbeat=workers.PrintWorkflow)
    item.root = True

    coordinator.input_queue.put(item)
    coordinator.wait_one()
    coordinator.stop()
    coordinator.join()


def main(argv):
    try:
        argv = FLAGS(argv)
    except gflags.FlagsError, e:
        print '%s\nUsage: %s ARGS\n%s' % (e, sys.argv[0], FLAGS)
        sys.exit(1)

    if len(argv) != 3:
        print 'Must supply two URLs as arguments.'
        sys.exit(1)

    assert FLAGS.upload_build_id
    assert FLAGS.release_server_prefix

    if FLAGS.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    real_main(
        new_url=argv[1],
        baseline_url=argv[2],
        upload_build_id=FLAGS.upload_build_id,
        upload_release_name=FLAGS.upload_release_name)


if __name__ == '__main__':
    main(sys.argv)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements authentication for the API server and frontend."""

import datetime
import functools
import json
import logging
import time
import urllib
import urllib2

# Local libraries
import flask
from flask import abort, g, redirect, render_template, request, url_for
from flask.ext.login import (
    confirm_login, current_user, fresh_login_required, login_fresh,
    login_required, login_user, logout_user)

# Local modules
from . import app
from . import db
from . import login
import config
from dpxdt.server import forms
from dpxdt.server import models
from dpxdt.server import operations
from dpxdt.server import utils

GOOGLE_OAUTH2_AUTH_URL = 'https://accounts.google.com/o/oauth2/auth'
GOOGLE_OAUTH2_TOKEN_URL = 'https://accounts.google.com/o/oauth2/token'
GOOGLE_OAUTH2_USERINFO_URL = 'https://www.googleapis.com/oauth2/v1/userinfo'
GOOGLE_OAUTH2_SCOPES ='https://www.googleapis.com/auth/userinfo.email'
FETCH_TIMEOUT_SECONDS = 60


@login.user_loader
def load_user(user_id):
    user = operations.UserOps(user_id).load()
    if user and user.is_authenticated():
        logging.debug('Authenticated as user=%r', user.get_id())
    return user


@app.context_processor
def auth_context():
    """Adds extra default context for rendered templates."""
    return dict(current_user=current_user)


@app.route('/login')
def login_view():
    next_url = request.args.get('next', default='/', type=str)

    if app.config.get('IGNORE_AUTH'):
        fake_id = 'anonymous_superuser'
        anonymous_superuser = models.User.query.get(fake_id)
        if not anonymous_superuser:
            anonymous_superuser = models.User(
                id=fake_id,
                email_address='superuser@example.com',
                superuser=1)
            db.session.add(anonymous_superuser)
            db.session.commit()
        login_user(anonymous_superuser)
        confirm_login()
        return redirect(next_url)

    # Inspired by:
    #   http://stackoverflow.com/questions/9499286
    #   /using-google-oauth2-with-flask
    params = dict(
        response_type='code',
        client_id=config.GOOGLE_OAUTH2_CLIENT_ID,
        redirect_uri=config.GOOGLE_OAUTH2_REDIRECT_URI,
        scope=GOOGLE_OAUTH2_SCOPES,
        state=urllib.quote(next_url),
    )
    target_url = '%s?%s' % (
        GOOGLE_OAUTH2_AUTH_URL, urllib.urlencode(params))
    logging.debug('Redirecting user to login at url=%r', target_url)
    return redirect(target_url)


@app.route('/logout')
@login_required
def logout():
    logout_user()
    return redirect(url_for('homepage'))


@app.route(config.GOOGLE_OAUTH2_REDIRECT_PATH)
def login_auth():
    # TODO: Handle when the 'error' parameter is present
    params = dict(
        code=request.args.get('code'),
        client_id=config.GOOGLE_OAUTH2_CLIENT_ID,
        client_secret=config.GOOGLE_OAUTH2_CLIENT_SECRET,
        redirect_uri=config.GOOGLE_OAUTH2_REDIRECT_URI,
        grant_type='authorization_code'
    )
    payload = urllib.urlencode(params)
    logging.debug('Posting for token to url=%r, payload=%r',
                  GOOGLE_OAUTH2_TOKEN_URL, payload)
    fetch_request = urllib2.Request(GOOGLE_OAUTH2_TOKEN_URL, payload)
    conn = urllib2.urlopen(fetch_request, timeout=FETCH_TIMEOUT_SECONDS)
    data = conn.read()
    result_dict = json.loads(data)

    params = dict(
        access_token=result_dict['access_token']
    )
    payload = urllib.urlencode(params)
    target_url = '%s?%s' % (GOOGLE_OAUTH2_USERINFO_URL, payload)
    logging.debug('Fetching user info from url=%r', target_url)
    fetch_request = urllib2.Request(target_url)
    conn = urllib2.urlopen(fetch_request, timeout=FETCH_TIMEOUT_SECONDS)
    data = conn.read()
    result_dict = json.loads(data)
    logging.debug('Result user info dict: %r', result_dict)
    email_address = result_dict['email']

    if not result_dict['verified_email']:
        abort(flask.Response('Your email address must be verified', 403))

    user_id = '%s:%s' % (models.User.GOOGLE_OAUTH2, result_dict['id'])
    user = models.User.query.get(user_id)
    if not user:
        user = models.User(id=user_id)

    # Email address on the account may change, user ID will stay the same.
    # Do not allow the user to claim existing build invitations with their
    # old email address.
    if user.email_address != email_address:
        user.email_address = email_address

    user.last_seen = datetime.datetime.utcnow()

    db.session.add(user)
    db.session.commit()

    login_user(user)
    confirm_login()

    # Clear all flashed messages from the session on login.
    flask.get_flashed_messages()

    final_url = urllib.unquote(request.args.get('state'))
    logging.debug('User is logged in. Redirecting to url=%r', final_url)
    return redirect(final_url)


@app.route('/whoami')
@login_required
def debug_login():
    return render_template(
        'whoami.html', user=current_user)


def superuser_required(f):
    """Requires the requestor to be a super user."""
    @functools.wraps(f)
    @login_required
    def wrapped(*args, **kwargs):
        if not (current_user.is_authenticated() and current_user.superuser):
            abort(403)
        return f(*args, **kwargs)
    return wrapped


def can_user_access_build(param_name):
    """Determines if the current user can access the build ID in the request.

    Args:
        param_name: Parameter name to use for getting the build ID from the
            request. Will fetch from GET or POST requests.

    Returns:
        The build the user has access to.
    """
    build_id = (
        request.args.get(param_name, type=int) or
        request.form.get(param_name, type=int))
    if not build_id:
        logging.debug('Build ID in param_name=%r was missing', param_name)
        abort(400)

    ops = operations.UserOps(current_user.get_id())
    build, user_is_owner = ops.owns_build(build_id)
    if not build:
        logging.debug('Could not find build_id=%r', build_id)
        abort(404)

    if current_user.is_authenticated() and not user_is_owner:
        # Assume the user should be able to access the build but can't because
        # the cache is out of date. This forces the cache to repopulate, any
        # outstanding user invitations to be completed, hopefully resulting in
        # the user having access to the build.
        ops.evict()
        claim_invitations(current_user)
        build, user_is_owner = ops.owns_build(build_id)

    if not user_is_owner:
        if current_user.is_authenticated() and current_user.superuser:
            pass
        elif request.method != 'GET':
            logging.debug('No way to log in user via modifying request')
            abort(403)
        elif build.public:
            pass
        elif current_user.is_authenticated():
            logging.debug('User does not have access to this build')
            abort(flask.Response('You cannot access this build', 403))
        else:
            logging.debug('Redirecting user to login to get build access')
            abort(login.unauthorized())
    elif not login_fresh():
        logging.debug('User login is old; forcing refresh')
        abort(login.needs_refresh())

    return build


def build_access_required(function_or_param_name):
    """Decorator ensures user has access to the build ID in the request.

    May be used in two ways:

        @build_access_required
        def my_func(build):
            ...

        @build_access_required('custom_build_id_param')
        def my_func(build):
            ...

    Always calls the given function with the models.Build entity as the
    first positional argument.
    """
    def get_wrapper(param_name, f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            g.build = can_user_access_build(param_name)
            if not utils.is_production():
                # Insert a sleep to emulate page loading in production.
                time.sleep(0.5)
            return f(*args, **kwargs)
        return wrapped

    if isinstance(function_or_param_name, basestring):
        return lambda f: get_wrapper(function_or_param_name, f)
    else:
        return get_wrapper('id', function_or_param_name)


def _get_api_key_ops():
    """Gets the operations.ApiKeyOps instance for the current request."""
    auth_header = request.authorization
    if not auth_header:
        logging.debug('API request lacks authorization header')
        abort(flask.Response(
            'API key required', 401,
            {'WWW-Authenticate': 'Basic realm="API key required"'}))

    return operations.ApiKeyOps(auth_header.username, auth_header.password)


def current_api_key():
    """Determines the API key for the current request.

    Returns:
        The ApiKey instance.
    """
    if app.config.get('IGNORE_AUTH'):
        return models.ApiKey(
            id='anonymous_superuser',
            secret='',
            superuser=True)

    ops = _get_api_key_ops()
    api_key = ops.get()
    logging.debug('Authenticated as API key=%r', api_key.id)

    return api_key


def can_api_key_access_build(param_name):
    """Determines if the current API key can access the build in the request.

    Args:
        param_name: Parameter name to use for getting the build ID from the
            request. Will fetch from GET or POST requests.

    Returns:
        (api_key, build) The API Key and the Build it has access to.
    """
    build_id = (
        request.args.get(param_name, type=int) or
        request.form.get(param_name, type=int))
    utils.jsonify_assert(build_id, 'build_id required')

    if app.config.get('IGNORE_AUTH'):
        api_key = models.ApiKey(
            id='anonymous_superuser',
            secret='',
            superuser=True)
        build = models.Build.query.get(build_id)
        utils.jsonify_assert(build is not None, 'build must exist', 404)
    else:
        ops = _get_api_key_ops()
        api_key, build = ops.can_access_build(build_id)

    return api_key, build


def build_api_access_required(f):
    """Decorator ensures API key has access to the build ID in the request.

    Always calls the given function with the models.Build entity as the
    first positional argument.
    """
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        g.api_key, g.build = can_api_key_access_build('build_id')
        return f(*args, **kwargs)
    return wrapped


def superuser_api_key_required(f):
    """Decorator ensures only superuser API keys can request this function."""
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        api_key = current_api_key()
        g.api_key = api_key

        utils.jsonify_assert(
            api_key.superuser,
            'API key=%r must be a super user' % api_key.id,
            403)

        return f(*args, **kwargs)

    return wrapped


@app.route('/api_keys', methods=['GET', 'POST'])
@fresh_login_required
@build_access_required('build_id')
def manage_api_keys():
    """Page for viewing and creating API keys."""
    build = g.build
    create_form = forms.CreateApiKeyForm()
    if create_form.validate_on_submit():
        api_key = models.ApiKey()
        create_form.populate_obj(api_key)
        api_key.id = utils.human_uuid()
        api_key.secret = utils.password_uuid()

        save_admin_log(build, created_api_key=True, message=api_key.id)

        db.session.add(api_key)
        db.session.commit()

        logging.info('Created API key=%r for build_id=%r',
                     api_key.id, build.id)
        return redirect(url_for('manage_api_keys', build_id=build.id))

    create_form.build_id.data = build.id

    api_key_query = (
        models.ApiKey.query
        .filter_by(build_id=build.id)
        .order_by(models.ApiKey.created.desc())
        .limit(1000))

    revoke_form_list = []
    for api_key in api_key_query:
        form = forms.RevokeApiKeyForm()
        form.id.data = api_key.id
        form.build_id.data = build.id
        form.revoke.data = True
        revoke_form_list.append((api_key, form))

    return render_template(
        'view_api_keys.html',
        build=build,
        create_form=create_form,
        revoke_form_list=revoke_form_list)


@app.route('/api_keys.revoke', methods=['POST'])
@fresh_login_required
@build_access_required('build_id')
def revoke_api_key():
    """Form submission handler for revoking API keys."""
    build = g.build
    form = forms.RevokeApiKeyForm()
    if form.validate_on_submit():
        api_key = models.ApiKey.query.get(form.id.data)
        if api_key.build_id != build.id:
            logging.debug('User does not have access to API key=%r',
                          api_key.id)
            abort(403)

        api_key.active = False
        save_admin_log(build, revoked_api_key=True, message=api_key.id)

        db.session.add(api_key)
        db.session.commit()

        ops = operations.ApiKeyOps(api_key.id, api_key.secret)
        ops.evict()

    return redirect(url_for('manage_api_keys', build_id=build.id))


def claim_invitations(user):
    """Claims any pending invitations for the given user's email address."""
    # See if there are any build invitations present for the user with this
    # email address. If so, replace all those invitations with the real user.
    invitation_user_id = '%s:%s' % (
        models.User.EMAIL_INVITATION, user.email_address)
    invitation_user = models.User.query.get(invitation_user_id)
    if invitation_user:
        invited_build_list = list(invitation_user.builds)
        if not invited_build_list:
            return

        db.session.add(user)
        logging.debug('Found %d build admin invitations for id=%r, user=%r',
                      len(invited_build_list), invitation_user_id, user)

        for build in invited_build_list:
            build.owners.remove(invitation_user)
            if not build.is_owned_by(user.id):
                build.owners.append(user)
                logging.debug('Claiming invitation for build_id=%r', build.id)
                save_admin_log(build, invite_accepted=True)
            else:
                logging.debug('User already owner of build. '
                              'id=%r, build_id=%r', user.id, build.id)
            db.session.add(build)

        db.session.delete(invitation_user)
        db.session.commit()
        # Re-add the user to the current session so we can query with it.
        db.session.add(current_user)


@app.route('/admins', methods=['GET', 'POST'])
@fresh_login_required
@build_access_required('build_id')
def manage_admins():
    """Page for viewing and managing build admins."""
    build = g.build

    # Do not show cached data
    db.session.add(build)
    db.session.refresh(build)

    add_form = forms.AddAdminForm()
    if add_form.validate_on_submit():

        invitation_user_id = '%s:%s' % (
            models.User.EMAIL_INVITATION, add_form.email_address.data)

        invitation_user = models.User.query.get(invitation_user_id)
        if not invitation_user:
            invitation_user = models.User(
                id=invitation_user_id,
                email_address=add_form.email_address.data)
            db.session.add(invitation_user)

        db.session.add(build)
        db.session.add(invitation_user)
        db.session.refresh(build, lockmode='update')

        build.owners.append(invitation_user)
        save_admin_log(build, invited_new_admin=True,
                       message=invitation_user.email_address)

        db.session.commit()

        logging.info('Added user=%r as owner to build_id=%r',
                     invitation_user.id, build.id)
        return redirect(url_for('manage_admins', build_id=build.id))

    add_form.build_id.data = build.id

    revoke_form_list = []
    for user in build.owners:
        form = forms.RemoveAdminForm()
        form.user_id.data = user.id
        form.build_id.data = build.id
        form.revoke.data = True
        revoke_form_list.append((user, form))

    return render_template(
        'view_admins.html',
        build=build,
        add_form=add_form,
        revoke_form_list=revoke_form_list)


@app.route('/admins.revoke', methods=['POST'])
@fresh_login_required
@build_access_required('build_id')
def revoke_admin():
    """Form submission handler for revoking admin access to a build."""
    build = g.build
    form = forms.RemoveAdminForm()
    if form.validate_on_submit():
        user = models.User.query.get(form.user_id.data)
        if not user:
            logging.debug('User being revoked admin access does not exist.'
                          'id=%r, build_id=%r', form.user_id.data, build.id)
            abort(400)

        if user == current_user:
            logging.debug('User trying to remove themself as admin. '
                          'id=%r, build_id=%r', user.id, build.id)
            abort(400)

        db.session.add(build)
        db.session.add(user)
        db.session.refresh(build, lockmode='update')
        db.session.refresh(user, lockmode='update')

        user_is_owner = build.owners.filter_by(id=user.id)
        if not user_is_owner:
            logging.debug('User being revoked admin access is not owner. '
                          'id=%r, build_id=%r.', user.id, build.id)
            abort(400)

        build.owners.remove(user)
        save_admin_log(build, revoked_admin=True, message=user.email_address)

        db.session.commit()

        operations.UserOps(user.get_id()).evict()

    return redirect(url_for('manage_admins', build_id=build.id))


def save_admin_log(build, **kwargs):
    """Saves an action to the admin log."""
    message = kwargs.pop('message', None)
    release = kwargs.pop('release', None)
    run = kwargs.pop('run', None)

    if not len(kwargs) == 1:
        raise TypeError('Must specify a LOG_TYPE argument')

    log_enum = kwargs.keys()[0]
    log_type = getattr(models.AdminLog, log_enum.upper(), None)
    if not log_type:
        raise TypeError('Bad log_type argument: %s' % log_enum)

    if current_user.is_anonymous():
        user_id = None
    else:
        user_id = current_user.get_id()

    log = models.AdminLog(
        build_id=build.id,
        log_type=log_type,
        message=message,
        user_id=user_id)

    if release:
        log.release_id = release.id

    if run:
        log.run_id = run.id
        log.release_id = run.release_id

    db.session.add(log)


@app.route('/activity')
@fresh_login_required
@build_access_required('build_id')
def view_admin_log():
    """Page for viewing the log of admin activity."""
    build = g.build

    # TODO: Add paging

    log_list = (
        models.AdminLog.query
        .filter_by(build_id=build.id)
        .order_by(models.AdminLog.created.desc())
        .all())

    return render_template(
        'view_admin_log.html',
        build=build,
        log_list=log_list)



#!/usr/bin/env python
# Copyright 2014 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility for diffing a set of URL pairs defined in a config file.

Example usage:

./dpxdt/tools/diff_my_urls.py \
    --upload_build_id=1234 \
    --release_server_prefix=https://my-dpxdt-apiserver.example.com/api \
    --release_client_id=<your api key> \
    --release_client_secret=<your api secret> \
    --upload_release_name="My release name" \
    --release_cut_url=http://example.com/path/to/my/release/tool/for/this/cut
    --tests_json_path=my_url_tests.json

Example input file "my_url_tests.json". One entry per test:

[
    {
        "name": "My homepage",
        "run_url": "http://localhost:5000/static/dummy/dummy_page1.html",
        "run_config": {
            "viewportSize": {
                "width": 1024,
                "height": 768
            },
            "injectCss": "#foobar { background-color: lime",
            "injectJs": "document.getElementById('foobar').innerText = 'bar';",
        },
        "ref_url": "http://localhost:5000/static/dummy/dummy_page1.html",
        "ref_config": {
            "viewportSize": {
                "width": 1024,
                "height": 768
            },
            "injectCss": "#foobar { background-color: goldenrod; }",
            "injectJs": "document.getElementById('foobar').innerText = 'foo';",
        }
    },
    ...
]

See README.md for documentation of config parameters.
"""

import datetime
import json
import logging
import sys

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

# Local modules
from dpxdt.client import fetch_worker
from dpxdt.client import release_worker
from dpxdt.client import workers
import flags


class Test(object):
    """Represents the JSON of a single test."""

    def __init__(self, name=None, run_url=None, run_config=None,
                 ref_url=None, ref_config=None):
        self.name = name
        self.run_url = run_url
        self.run_config_data = json.dumps(run_config)
        self.ref_url = ref_url
        self.ref_config_data = json.dumps(ref_config)


def load_tests(data):
    """Loads JSON data and returns a list of Test objects it contains."""
    test_list = json.loads(data)
    results = []
    for test_json in test_list:
        results.append(Test(**test_json))
    return results


class DiffMyUrls(workers.WorkflowItem):
    """Workflow for diffing a set of URL pairs defined in a config file.

    Args:
        release_url: URL of the newest and best version of the page.
        tests: List of Test objects to test.
        upload_build_id: Optional. Build ID of the site being compared. When
            supplied a new release will be cut for this build comparing it
            to the last good release.
        upload_release_name: Optional. Release name to use for the build. When
            not supplied, a new release based on the current time will be
            created.
        heartbeat: Function to call with progress status.
    """

    def run(self,
            release_url,
            tests,
            upload_build_id,
            upload_release_name,
            heartbeat=None):
        if not upload_release_name:
            upload_release_name = str(datetime.datetime.utcnow())

        yield heartbeat('Creating release %s' % upload_release_name)
        release_number = yield release_worker.CreateReleaseWorkflow(
            upload_build_id, upload_release_name, release_url)

        pending_uploads = []
        for test in tests:
            item = release_worker.RequestRunWorkflow(
                upload_build_id, upload_release_name, release_number,
                test.name, url=test.run_url, config_data=test.run_config_data,
                ref_url=test.ref_url, ref_config_data=test.ref_config_data)
            pending_uploads.append(item)

        yield heartbeat('Requesting %d runs' % len(pending_uploads))
        yield pending_uploads

        yield heartbeat('Marking runs as complete')
        release_url = yield release_worker.RunsDoneWorkflow(
            upload_build_id, upload_release_name, release_number)

        yield heartbeat('Results viewable at: %s' % release_url)


def real_main(release_url=None,
              tests_json_path=None,
              upload_build_id=None,
              upload_release_name=None):
    """Runs diff_my_urls."""
    coordinator = workers.get_coordinator()
    fetch_worker.register(coordinator)
    coordinator.start()

    data = open(FLAGS.tests_json_path).read()
    tests = load_tests(data)

    item = DiffMyUrls(
        release_url,
        tests,
        upload_build_id,
        upload_release_name,
        heartbeat=workers.PrintWorkflow)
    item.root = True

    coordinator.input_queue.put(item)
    coordinator.wait_one()
    coordinator.stop()
    coordinator.join()


def main(argv):
    try:
        argv = FLAGS(argv)
    except gflags.FlagsError, e:
        print '%s\nUsage: %s ARGS\n%s' % (e, sys.argv[0], FLAGS)
        sys.exit(1)

    assert FLAGS.release_cut_url
    assert FLAGS.release_server_prefix
    assert FLAGS.tests_json_path
    assert FLAGS.upload_build_id

    if FLAGS.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    real_main(
        release_url=FLAGS.release_cut_url,
        tests_json_path=FLAGS.tests_json_path,
        upload_build_id=FLAGS.upload_build_id,
        upload_release_name=FLAGS.upload_release_name)


if __name__ == '__main__':
    main(sys.argv)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pull-queue API."""

import datetime
import json
import logging
import time
import uuid

# Local modules
from . import app
from . import db
from dpxdt.server import signals


class Error(Exception):
    """Base class for exceptions in this module."""

class TaskDoesNotExistError(Error):
    """Task with the given ID does not exist and cannot be finished."""

class LeaseExpiredError(Error):
    """Owner's lease on the task has expired, not completing task."""

class NotOwnerError(Error):
    """Requestor is no longer the owner of the task."""


class WorkQueue(db.Model):
    """Represents a single item of work to do in a specific queue.

    Queries:
    - By task_id for finishing a task or extending a lease.
    - By Index(queue_name, status, eta) for finding the oldest task for a queue
        that is still pending.
    - By Index(status, create) for finding old tasks that should be deleted
        from the table periodically to free up space.
    """

    CANCELED = 'canceled'
    DONE = 'done'
    ERROR = 'error'
    LIVE = 'live'
    STATES = frozenset([CANCELED, DONE, ERROR, LIVE])

    task_id = db.Column(db.String(100), primary_key=True, nullable=False)
    queue_name = db.Column(db.String(100), primary_key=True, nullable=False)
    status = db.Column(db.Enum(*STATES), default=LIVE, nullable=False)
    eta = db.Column(db.DateTime, default=datetime.datetime.utcnow,
                    nullable=False)

    build_id = db.Column(db.Integer, db.ForeignKey('build.id'))
    release_id = db.Column(db.Integer, db.ForeignKey('release.id'))
    run_id = db.Column(db.Integer, db.ForeignKey('run.id'))

    source = db.Column(db.String(500))
    created = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    finished = db.Column(db.DateTime)

    lease_attempts = db.Column(db.Integer, default=0, nullable=False)
    last_lease = db.Column(db.DateTime)
    last_owner = db.Column(db.String(500))

    heartbeat = db.Column(db.Text)
    heartbeat_number = db.Column(db.Integer)

    payload = db.Column(db.LargeBinary)
    content_type = db.Column(db.String(100))

    __table_args__ = (
        db.Index('created_index', 'queue_name', 'status', 'created'),
        db.Index('lease_index', 'queue_name', 'status', 'eta'),
        db.Index('reap_index', 'status', 'created'),
    )

    @property
    def lease_outstanding(self):
        if not self.status == WorkQueue.LIVE:
            return False
        if not self.last_owner:
            return False
        now = datetime.datetime.utcnow()
        return now < self.eta


def add(queue_name, payload=None, content_type=None, source=None, task_id=None,
        build_id=None, release_id=None, run_id=None):
    """Adds a work item to a queue.

    Args:
        queue_name: Name of the queue to add the work item to.
        payload: Optional. Payload that describes the work to do as a string.
            If not a string and content_type is not provided, then this
            function assumes the payload is a JSON-able Python object.
        content_type: Optional. Content type of the payload.
        source: Optional. Who or what originally created the task.
        task_id: Optional. When supplied, only enqueue this task if a task
            with this ID does not already exist. If a task with this ID already
            exists, then this function will do nothing.
        build_id: Build ID to associate with this task. May be None.
        release_id: Release ID to associate with this task. May be None.
        run_id: Run ID to associate with this task. May be None.

    Returns:
        ID of the task that was added.
    """
    if task_id:
        task = WorkQueue.query.filter_by(task_id=task_id).first()
        if task:
            return task.task_id
    else:
        task_id = uuid.uuid4().hex

    if payload and not content_type and not isinstance(payload, basestring):
        payload = json.dumps(payload)
        content_type = 'application/json'

    now = datetime.datetime.utcnow()
    task = WorkQueue(
        task_id=task_id,
        queue_name=queue_name,
        eta=now,
        source=source,
        build_id=build_id,
        release_id=release_id,
        run_id=run_id,
        payload=payload,
        content_type=content_type)
    db.session.add(task)

    return task.task_id


def _datetime_to_epoch_seconds(dt):
    """Converts a datetime.datetime to seconds since the epoch."""
    if dt is None:
        return None
    return int(time.mktime(dt.utctimetuple()))


def _task_to_dict(task):
    """Converts a WorkQueue to a JSON-able dictionary."""
    payload = task.payload
    if payload and task.content_type == 'application/json':
        payload = json.loads(payload)

    return dict(
        task_id=task.task_id,
        queue_name=task.queue_name,
        eta=_datetime_to_epoch_seconds(task.eta),
        source=task.source,
        created=_datetime_to_epoch_seconds(task.created),
        lease_attempts=task.lease_attempts,
        last_lease=_datetime_to_epoch_seconds(task.last_lease),
        payload=payload,
        content_type=task.content_type)


# TODO: Allow requesting key to lease a task if the source matches. This
# would let users run their own workers for server-side capture queues.


def lease(queue_name, owner, count=1, timeout_seconds=60):
    """Leases a work item from a queue, usually the oldest task available.

    Args:
        queue_name: Name of the queue to lease work from.
        owner: Who or what is leasing the task.
        count: Lease up to this many tasks. Return value will never have more
            than this many items present.
        timeout_seconds: Number of seconds to lock the task for before
            allowing another owner to lease it.

    Returns:
        List of dictionaries representing the task that was leased, or
        an empty list if no tasks are available to be leased.
    """
    now = datetime.datetime.utcnow()
    query = (
        WorkQueue.query
        .filter_by(queue_name=queue_name, status=WorkQueue.LIVE)
        .filter(WorkQueue.eta <= now)
        .order_by(WorkQueue.eta)
        .with_lockmode('update')
        .limit(count))

    task_list = query.all()
    if not task_list:
        return None

    next_eta = now + datetime.timedelta(seconds=timeout_seconds)

    for task in task_list:
        task.eta = next_eta
        task.lease_attempts += 1
        task.last_owner = owner
        task.last_lease = now
        task.heartbeat = None
        task.heartbeat_number = 0
        db.session.add(task)

    return [_task_to_dict(task) for task in task_list]


def _get_task_with_policy(queue_name, task_id, owner):
    """Fetches the specified task and enforces ownership policy.

    Args:
        queue_name: Name of the queue the work item is on.
        task_id: ID of the task that is finished.
        owner: Who or what has the current lease on the task.

    Returns:
        The valid WorkQueue task that is currently owned.

    Raises:
        TaskDoesNotExistError if the task does not exist.
        LeaseExpiredError if the lease is no longer active.
        NotOwnerError if the specified owner no longer owns the task.
    """
    now = datetime.datetime.utcnow()
    task = (
        WorkQueue.query
        .filter_by(queue_name=queue_name, task_id=task_id)
        .with_lockmode('update')
        .first())
    if not task:
        raise TaskDoesNotExistError('task_id=%r' % task_id)

    # Lease delta should be positive, meaning it has not yet expired!
    lease_delta = now - task.eta
    if lease_delta > datetime.timedelta(0):
        db.session.rollback()
        raise LeaseExpiredError('queue=%r, task_id=%r expired %s' % (
                                task.queue_name, task_id, lease_delta))

    if task.last_owner != owner:
        db.session.rollback()
        raise NotOwnerError('queue=%r, task_id=%r, owner=%r' % (
                            task.queue_name, task_id, task.last_owner))

    return task


def heartbeat(queue_name, task_id, owner, message, index):
    """Sets the heartbeat status of the task and extends its lease.

    The task's lease is extended by the same amount as its last lease to
    ensure that any operations following the heartbeat will still hold the
    lock for the original lock period.

    Args:
        queue_name: Name of the queue the work item is on.
        task_id: ID of the task that is finished.
        owner: Who or what has the current lease on the task.
        message: Message to report as the task's current status.
        index: Number of this message in the sequence of messages from the
            current task owner, starting at zero. This lets the API receive
            heartbeats out of order, yet ensure that the most recent message
            is actually saved to the database. This requires the owner issuing
            heartbeat messages to issue heartbeat indexes sequentially.

    Returns:
        True if the heartbeat message was set, False if it is lower than the
        current heartbeat index.

    Raises:
        TaskDoesNotExistError if the task does not exist.
        LeaseExpiredError if the lease is no longer active.
        NotOwnerError if the specified owner no longer owns the task.
    """
    task = _get_task_with_policy(queue_name, task_id, owner)
    if task.heartbeat_number > index:
        return False

    task.heartbeat = message
    task.heartbeat_number = index

    # Extend the lease by the time of the last lease.
    now = datetime.datetime.utcnow()
    timeout_delta = task.eta - task.last_lease
    task.eta = now + timeout_delta
    task.last_lease = now

    db.session.add(task)

    signals.task_updated.send(app, task=task)

    return True


def finish(queue_name, task_id, owner, error=False):
    """Marks a work item on a queue as finished.

    Args:
        queue_name: Name of the queue the work item is on.
        task_id: ID of the task that is finished.
        owner: Who or what has the current lease on the task.
        error: Defaults to false. True if this task's final state is an error.

    Returns:
        True if the task has been finished for the first time; False if the
        task was already finished.

    Raises:
        TaskDoesNotExistError if the task does not exist.
        LeaseExpiredError if the lease is no longer active.
        NotOwnerError if the specified owner no longer owns the task.
    """
    task = _get_task_with_policy(queue_name, task_id, owner)

    if not task.status == WorkQueue.LIVE:
        logging.warning('Finishing already dead task. queue=%r, task_id=%r, '
                        'owner=%r, status=%r',
                        task.queue_name, task_id, owner, task.status)
        return False

    if not error:
        task.status = WorkQueue.DONE
    else:
        task.status = WorkQueue.ERROR

    task.finished = datetime.datetime.utcnow()
    db.session.add(task)

    signals.task_updated.send(app, task=task)

    return True


def _query(queue_name=None, build_id=None, release_id=None, run_id=None,
           count=None):
    """Queries for work items based on their criteria.

    Args:
        queue_name: Optional queue name to restrict to.
        build_id: Optional build ID to restrict to.
        release_id: Optional release ID to restrict to.
        run_id: Optional run ID to restrict to.
        count: How many tasks to fetch. Defaults to None, which means all
            tasks are fetch that match the query.

    Returns:
        List of WorkQueue items.
    """
    assert queue_name or build_id or release_id or run_id

    q = WorkQueue.query
    if queue_name:
        q = q.filter_by(queue_name=queue_name)
    if build_id:
        q = q.filter_by(build_id=build_id)
    if release_id:
        q = q.filter_by(release_id=release_id)
    if run_id:
        q = q.filter_by(run_id=run_id)

    q = q.order_by(WorkQueue.created.desc())

    if count is not None:
        q = q.limit(count)

    return q.all()


def query(**kwargs):
    """Queries for work items based on their criteria.

    Args:
        queue_name: Optional queue name to restrict to.
        build_id: Optional build ID to restrict to.
        release_id: Optional release ID to restrict to.
        run_id: Optional run ID to restrict to.
        count: How many tasks to fetch. Defaults to None, which means all
            tasks are fetch that match the query.

    Returns:
        Dictionaries of the most recent tasks that match the criteria, in
        order of most recently created. When count is 1 the return value will
        be the most recent task or None. When count is not 1 the return value
        will be a  list of tasks.
    """
    count = kwargs.get('count', None)
    task_list = _query(**kwargs)
    task_dict_list = [_task_to_dict(task) for task in task_list]

    if count == 1:
        if not task_dict_list:
            return None
        else:
            return task_dict_list[0]

    return task_dict_list


def cancel(**kwargs):
    """Cancels work items based on their criteria.

    Args:
        **kwargs: Same parameters as the query() method.

    Returns:
        The number of tasks that were canceled.
    """
    task_list = _query(**kwargs)
    for task in task_list:
        task.status = WorkQueue.CANCELED
        task.finished = datetime.datetime.utcnow()
        db.session.add(task)
    return len(task_list)



#!/usr/bin/env python
# Copyright 2014 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility for uploading and diffing images that were generated locally.

Plugs screenshots generated in a tool like Selenium into Depicted. Uses the
last known good screenshots for tests with the same name as the baseline for
comparison. Depicted will generate diffs for you and manage the workflow.

Example usage:

./dpxdt/tools/diff_my_images.py \
    --upload_build_id=1234 \
    --release_server_prefix=https://my-dpxdt-apiserver.example.com/api \
    --release_client_id=<your api key> \
    --release_client_secret=<your api secret> \
    --upload_release_name="My release name" \
    --release_cut_url=http://example.com/path/to/my/release/tool/for/this/cut
    --tests_json_path=my_tests.json

Example input file "my_tests.json". One entry per test:

[
    {
        "name": "My homepage",
        "run_failed": false,
        "image_path": "/tmp/path/to/my/new_screenshot.png",
        "log_path": "/tmp/path/to/my/new_output_log.txt",
        "url": "http://example.com/new/build/url/here"
    },
    ...
]

Use the "run_failed" parameter when your screenshotting tool failed for
some reason and you want to upload your log but still mark the test as
having failed. This makes it easy to debug all of your Depicted tests in
one place for a single release.
"""

import datetime
import json
import logging
import sys

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

# Local modules
from dpxdt.client import fetch_worker
from dpxdt.client import release_worker
from dpxdt.client import workers
import flags


class Test(object):
    """Represents the JSON of a single test."""

    def __init__(self, name=None, run_failed=False, image_path=None,
                 log_path=None, url=None):
        self.name = name
        self.run_failed = run_failed
        self.image_path = image_path
        self.log_path = log_path
        self.url = url


def load_tests(data):
    """Loads JSON data and returns a list of Test objects it contains."""
    test_list = json.loads(data)
    results = []
    for test_json in test_list:
        results.append(Test(**test_json))
    return results


class RunTest(workers.WorkflowItem):
    """Workflow to run a single test.

    Searches for the last good run for the same test name to use as a
    baseline for comparison. If no last good run is found, the supplied
    images will be treated as a new basline.

    Args:
        build_id: ID of the build being tested.
        release_name: Name of the release being tested.
        release_number: Number of the release being tested.
        test: Test object to handle.
        heartbeat: Function to call with progress status.
    """

    def run(self, build_id, release_name, release_number, test, heartbeat=None):
        ref_image, ref_log, ref_url = None, None, None
        try:
            last_good = yield release_worker.FindRunWorkflow(
                build_id, test.name)
        except release_worker.FindRunError, e:
            yield heartbeat('Could not find last good run for %s' % test.name)
        else:
            ref_image = last_good['image'] or None
            ref_log = last_good['log'] or None
            ref_url = last_good['url'] or None

        yield heartbeat('Uploading data for %s' % test.name)
        yield release_worker.ReportRunWorkflow(
            build_id,
            release_name,
            release_number,
            test.name,
            image_path=test.image_path,
            log_path=test.log_path,
            url=test.url,
            ref_image=ref_image,
            ref_log=ref_log,
            ref_url=ref_url,
            run_failed=test.run_failed)


class DiffMyImages(workers.WorkflowItem):
    """Workflow for diffing set of images generated outside of Depicted.

    Args:
        release_url: URL of the newest and best version of the page.
        tests: List of Test objects to test.
        upload_build_id: Optional. Build ID of the site being compared. When
            supplied a new release will be cut for this build comparing it
            to the last good release.
        upload_release_name: Optional. Release name to use for the build. When
            not supplied, a new release based on the current time will be
            created.
        heartbeat: Function to call with progress status.
    """

    def run(self,
            release_url,
            tests,
            upload_build_id,
            upload_release_name,
            heartbeat=None):
        if not upload_release_name:
            upload_release_name = str(datetime.datetime.utcnow())

        yield heartbeat('Creating release %s' % upload_release_name)
        release_number = yield release_worker.CreateReleaseWorkflow(
            upload_build_id, upload_release_name, release_url)

        pending_uploads = []
        for test in tests:
            item = RunTest(upload_build_id, upload_release_name,
                           release_number, test, heartbeat=heartbeat)
            pending_uploads.append(item)

        yield heartbeat('Uploading %d runs' % len(pending_uploads))
        yield pending_uploads

        yield heartbeat('Marking runs as complete')
        release_url = yield release_worker.RunsDoneWorkflow(
            upload_build_id, upload_release_name, release_number)

        yield heartbeat('Results viewable at: %s' % release_url)


def real_main(release_url=None,
              tests_json_path=None,
              upload_build_id=None,
              upload_release_name=None):
    """Runs diff_my_images."""
    coordinator = workers.get_coordinator()
    fetch_worker.register(coordinator)
    coordinator.start()

    data = open(FLAGS.tests_json_path).read()
    tests = load_tests(data)

    item = DiffMyImages(
        release_url,
        tests,
        upload_build_id,
        upload_release_name,
        heartbeat=workers.PrintWorkflow)
    item.root = True

    coordinator.input_queue.put(item)
    coordinator.wait_one()
    coordinator.stop()
    coordinator.join()


def main(argv):
    try:
        argv = FLAGS(argv)
    except gflags.FlagsError, e:
        print '%s\nUsage: %s ARGS\n%s' % (e, sys.argv[0], FLAGS)
        sys.exit(1)

    assert FLAGS.release_cut_url
    assert FLAGS.release_server_prefix
    assert FLAGS.tests_json_path
    assert FLAGS.upload_build_id

    if FLAGS.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    real_main(
        release_url=FLAGS.release_cut_url,
        tests_json_path=FLAGS.tests_json_path,
        upload_build_id=FLAGS.upload_build_id,
        upload_release_name=FLAGS.upload_release_name)


if __name__ == '__main__':
    main(sys.argv)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Signals for frontend and API server requests."""


# Local modules
from blinker import Namespace


_signals = Namespace()


# The build settings have been modified. Sender is the app. Arguments
# are (models.Build, models.User). Signal is sent immediately *after* the
# Build is committed to the DB.
build_updated = _signals.signal('build-updated')


# A release has been created or updated via the API. Sender is the app.
# Arguments are (models.Build, models.Release). Signal is sent immediately
# *after* the Release is committed to the DB.
release_updated_via_api = _signals.signal('release-update')


# A run has been created or updated via the API. Sender is the app. Arguments
# are (models.Build, models.Release, models.Run). Signal is sent immediately
# *after* the Run is committed to the DB.
run_updated_via_api = _signals.signal('run-updated')

# A WorkQueue task's status has been updated via the API. Sender is the app.
# Argument is (work_queue.WorkQueue). Signal is sent immediately after the
# task is updated but before it is committed to the DB.
task_updated = _signals.signal('task-updated')



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Web-based API for managing screenshots and incremental perceptual diffs.

Lifecycle of a release:

1. User creates a new build, which represents a single product or site that
   will be screenshotted repeatedly over time. This may happen very
   infrequenty through a web UI.

2. User creates a new release candidate for the build with a specific release
   name. The candidate is an attempt at finishing a specific release name. It
   may take many attempts, many candidates, before the release with that name
   is complete and can be marked as good.

3. User creates many runs for the candidate created in #2. Each run is
   identified by a unique name that describes what it does. For example, the
   run name could be the URL path for a page being screenshotted. The user
   associates each run with a new screenshot artifact. Runs are automatically
   associated with a corresponding run from the last good release. This makes
   it easy to compare new and old screenshots for runs with the same name.

4. User uploads a series of screenshot artifacts identified by content hash.
   Perceptual diffs between these new screenshots and the last good release
   may also be uploaded as an optimization. This may happen in parallel
   with #3.

5. The user marks the release candidate as having all of its expected runs
   present, meaning it will no longer receive new runs. This should only
   happen after all screenshot artifacts have finished uploading.

6. If a run indicates a previous screenshot, but no perceptual diff has
   been made to compare the new and old versions, a worker will do a perceptual
   diff, upload it, and associate it with the run.

7. Once all perceptual diffs for a release candidate's runs are complete,
   the results of the candidate are emailed out to the build's owner.

8. The build owner can go into a web UI, inspect the new/old perceptual diffs,
   and mark certain runs as okay even though the perceptual diff showed a
   difference. For example, a new feature will cause a perceptual diff, but
   should not be treated as a failure.

9. The user decides the release candidate looks correct and marks it as good,
   or the user thinks the candidate looks bad and goes back to #2 and begins
   creating a new candidate for that release all over again.


Notes:

- At any time, a user can manually mark any candidate or release as bad. This
  is useful to deal with bugs in the screenshotter, mistakes in approving a
  release candidate, rolling back to an earlier version, etc.

- As soon as a new release name is cut for a build, the last candidate of
  the last release is marked as good if there is no other good candidate. This
  lets the API establish a "baseline" release easily for first-time users.

- Only one release candidate may be receiving runs for a build at a time.

- Failure status can be indicated for a run at the capture phase or the
  diff phase. The API assumes that the same user that indicated the failure
  will also provide a log for the failing process so it can be inspected
  manually for a root cause. Uploading image artifacts for failed runs is
  not supported.
"""

import datetime
import hashlib
import functools
import json
import logging
import mimetypes
import time

# Local libraries
import flask
from flask import Flask, abort, g, request, url_for
from flask.exceptions import HTTPException

# Local modules
from . import app
from . import db
from dpxdt import constants
from dpxdt.server import auth
from dpxdt.server import emails
from dpxdt.server import models
from dpxdt.server import signals
from dpxdt.server import work_queue
from dpxdt.server import utils


@app.route('/api/create_release', methods=['POST'])
@auth.build_api_access_required
@utils.retryable_transaction()
def create_release():
    """Creates a new release candidate for a build."""
    build = g.build
    release_name = request.form.get('release_name')
    utils.jsonify_assert(release_name, 'release_name required')
    url = request.form.get('url')
    utils.jsonify_assert(release_name, 'url required')

    release = models.Release(
        name=release_name,
        url=url,
        number=1,
        build_id=build.id)

    last_candidate = (
        models.Release.query
        .filter_by(build_id=build.id, name=release_name)
        .order_by(models.Release.number.desc())
        .first())
    if last_candidate:
        release.number += last_candidate.number
        if last_candidate.status == models.Release.PROCESSING:
            canceled_task_count = work_queue.cancel(
                release_id=last_candidate.id)
            logging.info('Canceling %d tasks for previous attempt '
                         'build_id=%r, release_name=%r, release_number=%d',
                         canceled_task_count, build.id, last_candidate.name,
                         last_candidate.number)
            last_candidate.status = models.Release.BAD
            db.session.add(last_candidate)

    db.session.add(release)
    db.session.commit()

    signals.release_updated_via_api.send(app, build=build, release=release)

    logging.info('Created release: build_id=%r, release_name=%r, url=%r, '
                 'release_number=%d', build.id, release.name,
                 url, release.number)

    return flask.jsonify(
        success=True,
        build_id=build.id,
        release_name=release.name,
        release_number=release.number,
        url=url)


def _check_release_done_processing(release):
    """Moves a release candidate to reviewing if all runs are done."""
    if release.status != models.Release.PROCESSING:
        # NOTE: This statement also guards for situations where the user has
        # prematurely specified that the release is good or bad. Once the user
        # has done that, the system will not automatically move the release
        # back into the 'reviewing' state or send the email notification below.
        logging.info('Release not in processing state yet: build_id=%r, '
                     'name=%r, number=%d', release.build_id, release.name,
                     release.number)
        return False

    query = models.Run.query.filter_by(release_id=release.id)
    for run in query:
        if run.status == models.Run.NEEDS_DIFF:
            # Still waiting for the diff to finish.
            return False
        if run.ref_config and not run.ref_image:
            # Still waiting for the ref capture to process.
            return False
        if run.config and not run.image:
            # Still waiting for the run capture to process.
            return False

    logging.info('Release done processing, now reviewing: build_id=%r, '
                 'name=%r, number=%d', release.build_id, release.name,
                 release.number)

    # Send the email at the end of this request so we know it's only
    # sent a single time (guarded by the release.status check above).
    build_id = release.build_id
    release_name = release.name
    release_number = release.number

    @utils.after_this_request
    def send_notification_email(response):
        emails.send_ready_for_review(build_id, release_name, release_number)

    release.status = models.Release.REVIEWING
    db.session.add(release)
    return True


def _get_release_params():
    """Gets the release params from the current request."""
    release_name = request.form.get('release_name')
    utils.jsonify_assert(release_name, 'release_name required')
    release_number = request.form.get('release_number', type=int)
    utils.jsonify_assert(release_number is not None, 'release_number required')
    return release_name, release_number


def _find_last_good_run(build):
    """Finds the last good release and run for a build."""
    run_name = request.form.get('run_name', type=str)
    utils.jsonify_assert(run_name, 'run_name required')

    last_good_release = (
        models.Release.query
        .filter_by(
            build_id=build.id,
            status=models.Release.GOOD)
        .order_by(models.Release.created.desc())
        .first())

    last_good_run = None

    if last_good_release:
        logging.debug('Found last good release for: build_id=%r, '
                      'release_name=%r, release_number=%d',
                      build.id, last_good_release.name,
                      last_good_release.number)
        last_good_run = (
            models.Run.query
            .filter_by(release_id=last_good_release.id, name=run_name)
            .first())
        if last_good_run:
            logging.debug('Found last good run for: build_id=%r, '
                          'release_name=%r, release_number=%d, '
                          'run_name=%r',
                          build.id, last_good_release.name,
                          last_good_release.number, last_good_run.name)

    return last_good_release, last_good_run


@app.route('/api/find_run', methods=['POST'])
@auth.build_api_access_required
def find_run():
    """Finds the last good run of the given name for a release."""
    build = g.build
    last_good_release, last_good_run = _find_last_good_run(build)

    if last_good_run:
        return flask.jsonify(
            success=True,
            build_id=build.id,
            release_name=last_good_release.name,
            release_number=last_good_release.number,
            run_name=last_good_run.name,
            url=last_good_run.url,
            image=last_good_run.image,
            log=last_good_run.log,
            config=last_good_run.config)

    return utils.jsonify_error('Run not found')


def _get_or_create_run(build):
    """Gets a run for a build or creates it if it does not exist."""
    release_name, release_number = _get_release_params()
    run_name = request.form.get('run_name', type=str)
    utils.jsonify_assert(run_name, 'run_name required')

    release = (
        models.Release.query
        .filter_by(build_id=build.id, name=release_name, number=release_number)
        .first())
    utils.jsonify_assert(release, 'release does not exist')

    run = (
        models.Run.query
        .filter_by(release_id=release.id, name=run_name)
        .first())
    if not run:
        # Ignore re-reports of the same run name for this release.
        logging.info('Created run: build_id=%r, release_name=%r, '
                     'release_number=%d, run_name=%r',
                     build.id, release.name, release.number, run_name)
        run = models.Run(
            release_id=release.id,
            name=run_name,
            status=models.Run.DATA_PENDING)
        db.session.add(run)
        db.session.flush()

    return release, run


def _enqueue_capture(build, release, run, url, config_data, baseline=False):
    """Enqueues a task to run a capture process."""
    # Validate the JSON config parses.
    try:
        config_dict = json.loads(config_data)
    except Exception, e:
        abort(utils.jsonify_error(e))

    # Rewrite the config JSON to include the URL specified in this request.
    # Blindly overwrite anything that was there.
    config_dict['targetUrl'] = url
    config_data = json.dumps(config_dict)

    config_artifact = _save_artifact(build, config_data, 'application/json')
    db.session.add(config_artifact)
    db.session.flush()

    suffix = ''
    if baseline:
        suffix = ':baseline'

    task_id = '%s:%s%s' % (run.id, hashlib.sha1(url).hexdigest(), suffix)
    logging.info('Enqueueing capture task=%r, baseline=%r', task_id, baseline)

    work_queue.add(
        constants.CAPTURE_QUEUE_NAME,
        payload=dict(
            build_id=build.id,
            release_name=release.name,
            release_number=release.number,
            run_name=run.name,
            url=url,
            config_sha1sum=config_artifact.id,
            baseline=baseline,
        ),
        build_id=build.id,
        release_id=release.id,
        run_id=run.id,
        source='request_run',
        task_id=task_id)

    # Set the URL and config early to indicate to report_run that there is
    # still data pending even if 'image' and 'ref_image' are unset.
    if baseline:
        run.ref_url = url
        run.ref_config = config_artifact.id
    else:
        run.url = url
        run.config = config_artifact.id


@app.route('/api/request_run', methods=['POST'])
@auth.build_api_access_required
@utils.retryable_transaction()
def request_run():
    """Requests a new run for a release candidate."""
    build = g.build
    current_release, current_run = _get_or_create_run(build)

    current_url = request.form.get('url', type=str)
    config_data = request.form.get('config', default='{}', type=str)
    utils.jsonify_assert(current_url, 'url to capture required')
    utils.jsonify_assert(config_data, 'config document required')

    config_artifact = _enqueue_capture(
        build, current_release, current_run, current_url, config_data)

    ref_url = request.form.get('ref_url', type=str)
    ref_config_data = request.form.get('ref_config', type=str)
    utils.jsonify_assert(
        bool(ref_url) == bool(ref_config_data),
        'ref_url and ref_config must both be specified or not specified')

    if ref_url and ref_config_data:
        ref_config_artifact = _enqueue_capture(
            build, current_release, current_run, ref_url, ref_config_data,
            baseline=True)
    else:
        _, last_good_run = _find_last_good_run(build)
        if last_good_run:
            current_run.ref_url = last_good_run.url
            current_run.ref_image = last_good_run.image
            current_run.ref_log = last_good_run.log
            current_run.ref_config = last_good_run.config

    db.session.add(current_run)
    db.session.commit()

    signals.run_updated_via_api.send(
        app, build=build, release=current_release, run=current_run)

    return flask.jsonify(
        success=True,
        build_id=build.id,
        release_name=current_release.name,
        release_number=current_release.number,
        run_name=current_run.name,
        url=current_run.url,
        config=current_run.config,
        ref_url=current_run.ref_url,
        ref_config=current_run.ref_config)


@app.route('/api/report_run', methods=['POST'])
@auth.build_api_access_required
@utils.retryable_transaction()
def report_run():
    """Reports data for a run for a release candidate."""
    build = g.build
    release, run = _get_or_create_run(build)

    db.session.refresh(run, lockmode='update')

    current_url = request.form.get('url', type=str)
    current_image = request.form.get('image', type=str)
    current_log = request.form.get('log', type=str)
    current_config = request.form.get('config', type=str)

    ref_url = request.form.get('ref_url', type=str)
    ref_image = request.form.get('ref_image', type=str)
    ref_log = request.form.get('ref_log', type=str)
    ref_config = request.form.get('ref_config', type=str)

    diff_failed = request.form.get('diff_failed', type=str)
    diff_image = request.form.get('diff_image', type=str)
    diff_log = request.form.get('diff_log', type=str)

    distortion = request.form.get('distortion', default=None, type=float)
    run_failed = request.form.get('run_failed', type=str)

    if current_url:
        run.url = current_url
    if current_image:
        run.image = current_image
    if current_log:
        run.log = current_log
    if current_config:
        run.config = current_config
    if current_image or current_log or current_config:
        logging.info('Saving run data: build_id=%r, release_name=%r, '
                     'release_number=%d, run_name=%r, url=%r, '
                     'image=%r, log=%r, config=%r, run_failed=%r',
                     build.id, release.name, release.number, run.name,
                     run.url, run.image, run.log, run.config, run_failed)

    if ref_url:
        run.ref_url = ref_url
    if ref_image:
        run.ref_image = ref_image
    if ref_log:
        run.ref_log = ref_log
    if ref_config:
        run.ref_config = ref_config
    if ref_image or ref_log or ref_config:
        logging.info('Saved reference data: build_id=%r, release_name=%r, '
                     'release_number=%d, run_name=%r, ref_url=%r, '
                     'ref_image=%r, ref_log=%r, ref_config=%r',
                     build.id, release.name, release.number, run.name,
                     run.ref_url, run.ref_image, run.ref_log, run.ref_config)

    if diff_image:
        run.diff_image = diff_image
    if diff_log:
        run.diff_log = diff_log
    if distortion:
        run.distortion = distortion

    if diff_image or diff_log:
        logging.info('Saved pdiff: build_id=%r, release_name=%r, '
                     'release_number=%d, run_name=%r, diff_image=%r, '
                     'diff_log=%r, diff_failed=%r, distortion=%r',
                     build.id, release.name, release.number, run.name,
                     run.diff_image, run.diff_log, diff_failed, distortion)

    if run.image and run.diff_image:
        run.status = models.Run.DIFF_FOUND
    elif run.image and run.ref_image and not run.diff_log:
        run.status = models.Run.NEEDS_DIFF
    elif run.image and run.ref_image and not diff_failed:
        run.status = models.Run.DIFF_NOT_FOUND
    elif run.image and not run.ref_config:
        run.status = models.Run.NO_DIFF_NEEDED
    elif run_failed or diff_failed:
        run.status = models.Run.FAILED
    else:
        # NOTE: Intentionally do not transition state here in the default case.
        # We allow multiple background workers to be writing to the same Run in
        # parallel updating its various properties.
        pass

    # TODO: Verify the build has access to both the current_image and
    # the reference_sha1sum so they can't make a diff from a black image
    # and still see private data in the diff image.

    if run.status == models.Run.NEEDS_DIFF:
        task_id = '%s:%s:%s' % (run.id, run.image, run.ref_image)
        logging.info('Enqueuing pdiff task=%r', task_id)

        work_queue.add(
            constants.PDIFF_QUEUE_NAME,
            payload=dict(
                build_id=build.id,
                release_name=release.name,
                release_number=release.number,
                run_name=run.name,
                run_sha1sum=run.image,
                reference_sha1sum=run.ref_image,
            ),
            build_id=build.id,
            release_id=release.id,
            run_id=run.id,
            source='report_run',
            task_id=task_id)

    # Flush the run so querying for Runs in _check_release_done_processing
    # will be find the new run too and we won't deadlock.
    db.session.add(run)
    db.session.flush()

    _check_release_done_processing(release)
    db.session.commit()

    signals.run_updated_via_api.send(
        app, build=build, release=release, run=run)

    logging.info('Updated run: build_id=%r, release_name=%r, '
                 'release_number=%d, run_name=%r, status=%r',
                 build.id, release.name, release.number, run.name, run.status)

    return flask.jsonify(success=True)


@app.route('/api/runs_done', methods=['POST'])
@auth.build_api_access_required
@utils.retryable_transaction()
def runs_done():
    """Marks a release candidate as having all runs reported."""
    build = g.build
    release_name, release_number = _get_release_params()

    release = (
        models.Release.query
        .filter_by(build_id=build.id, name=release_name, number=release_number)
        .with_lockmode('update')
        .first())
    utils.jsonify_assert(release, 'Release does not exist')

    release.status = models.Release.PROCESSING
    db.session.add(release)
    _check_release_done_processing(release)
    db.session.commit()

    signals.release_updated_via_api.send(app, build=build, release=release)

    logging.info('Runs done for release: build_id=%r, release_name=%r, '
                 'release_number=%d', build.id, release.name, release.number)

    results_url = url_for(
        'view_release',
        id=build.id,
        name=release.name,
        number=release.number,
        _external=True)

    return flask.jsonify(
        success=True,
        results_url=results_url)


def _artifact_created(artifact):
    """Called whenever an Artifact is created.

    This method may be overridden in environments that have a different way of
    storing artifact files, such as on-disk or S3. Use the artifact.alternate
    field to hold the environment-specific data you need.
    """
    pass


def _save_artifact(build, data, content_type):
    """Saves an artifact to the DB and returns it."""
    sha1sum = hashlib.sha1(data).hexdigest()
    artifact = models.Artifact.query.filter_by(id=sha1sum).first()

    if artifact:
      logging.debug('Upload already exists: artifact_id=%r', sha1sum)
    else:
      logging.info('Upload received: artifact_id=%r, content_type=%r',
                   sha1sum, content_type)
      artifact = models.Artifact(
          id=sha1sum,
          content_type=content_type,
          data=data)
      _artifact_created(artifact)

    artifact.owners.append(build)
    return artifact


@app.route('/api/upload', methods=['POST'])
@auth.build_api_access_required
@utils.retryable_transaction()
def upload():
    """Uploads an artifact referenced by a run."""
    build = g.build
    utils.jsonify_assert(len(request.files) == 1,
                         'Need exactly one uploaded file')

    file_storage = request.files.values()[0]
    data = file_storage.read()
    content_type, _ = mimetypes.guess_type(file_storage.filename)

    artifact = _save_artifact(build, data, content_type)

    db.session.add(artifact)
    db.session.commit()

    return flask.jsonify(
        success=True,
        build_id=build.id,
        sha1sum=artifact.id,
        content_type=content_type)


def _get_artifact_response(artifact):
    """Gets the response object for the given artifact.

    This method may be overridden in environments that have a different way of
    storing artifact files, such as on-disk or S3.
    """
    response = flask.Response(
        artifact.data,
        mimetype=artifact.content_type)
    response.cache_control.public = True
    response.cache_control.max_age = 8640000
    response.set_etag(artifact.id)
    return response


@app.route('/api/download')
def download():
    """Downloads an artifact by it's content hash."""
    # Allow users with access to the build to download the file. Falls back
    # to API keys with access to the build. Prefer user first for speed.
    try:
        build = auth.can_user_access_build('build_id')
    except HTTPException:
        logging.debug('User access to artifact failed. Trying API key.')
        _, build = auth.can_api_key_access_build('build_id')

    sha1sum = request.args.get('sha1sum', type=str)
    if not sha1sum:
        logging.debug('Artifact sha1sum=%r not supplied', sha1sum)
        abort(404)

    artifact = models.Artifact.query.get(sha1sum)
    if not artifact:
        logging.debug('Artifact sha1sum=%r does not exist', sha1sum)
        abort(404)

    build_id = request.args.get('build_id', type=int)
    if not build_id:
        logging.debug('build_id missing for artifact sha1sum=%r', sha1sum)
        abort(404)

    is_owned = artifact.owners.filter_by(id=build_id).first()
    if not is_owned:
        logging.debug('build_id=%r not owner of artifact sha1sum=%r',
                      build_id, sha1sum)
        abort(403)

    # Make sure there are no Set-Cookie headers on the response so this
    # request is cachable by all HTTP frontends.
    @utils.after_this_request
    def no_session(response):
        if 'Set-Cookie' in response.headers:
            del response.headers['Set-Cookie']

    if not utils.is_production():
        # Insert a sleep to emulate how the page loading looks in production.
        time.sleep(1.5)

    if request.if_none_match and request.if_none_match.contains(sha1sum):
        response = flask.Response(status=304)
        return response

    return _get_artifact_response(artifact)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Forms for parsing and validating frontend requests."""

import datetime

# Local libraries
from flask.ext.wtf import (
    BooleanField, DataRequired, Email, Form, HiddenField, IntegerField,
    Length, NumberRange, Optional, Required, SubmitField, TextField)

# Local modules
from . import app


class BuildForm(Form):
    """Form for creating or editing a build."""

    name = TextField(validators=[Length(min=1, max=200)])


class ReleaseForm(Form):
    """Form for viewing or approving a release."""

    id = HiddenField(validators=[NumberRange(min=1)])
    name = HiddenField(validators=[Length(min=1, max=200)])
    number = HiddenField(validators=[NumberRange(min=1)])

    good = HiddenField()
    bad = HiddenField()
    reviewing = HiddenField()


class RunForm(Form):
    """Form for viewing or approving a run."""

    id = HiddenField(validators=[NumberRange(min=1)])
    name = HiddenField(validators=[Length(min=1, max=200)])
    number = HiddenField(validators=[NumberRange(min=1)])
    test = HiddenField(validators=[Length(min=1, max=200)])
    type = HiddenField(validators=[Length(min=1, max=200)])
    approve = HiddenField()
    disapprove = HiddenField()


class CreateApiKeyForm(Form):
    """Form for creating an API key."""

    build_id = HiddenField(validators=[NumberRange(min=1)])
    purpose = TextField('Purpose', validators=[Length(min=1, max=200)])
    create = SubmitField('Create')


class RevokeApiKeyForm(Form):
    """Form for revoking an API key."""

    id = HiddenField()
    build_id = HiddenField(validators=[NumberRange(min=1)])
    revoke = SubmitField('Revoke')


class AddAdminForm(Form):
    """Form for adding a build admin."""

    email_address = TextField('Email address',
                              validators=[Length(min=1, max=200)])
    build_id = HiddenField(validators=[NumberRange(min=1)])
    add = SubmitField('Add')


class RemoveAdminForm(Form):
    """Form for removing a build admin."""

    user_id = HiddenField(validators=[Length(min=1, max=200)])
    build_id = HiddenField(validators=[NumberRange(min=1)])
    revoke = SubmitField('Revoke')


class ModifyWorkQueueTaskForm(Form):
    """Form for modifying a work queue task."""

    task_id = HiddenField()
    action = HiddenField()
    delete = SubmitField('Delete')
    retry = SubmitField('Retry')


class SettingsForm(Form):
    """Form for modifying build settings."""

    name = TextField(validators=[Length(min=1, max=200)])
    send_email = BooleanField('Send notification emails')
    email_alias = TextField('Mailing list for notifications',
                            validators=[Optional(), Email()])
    build_id = HiddenField(validators=[NumberRange(min=1)])
    save = SubmitField('Save')



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common utility functions."""

import base64
import datetime
import hashlib
import functools
import logging
import os
import traceback
import uuid

# Local libraries
import flask
from flask import abort, g, jsonify
from sqlalchemy.exc import OperationalError

# Local modules
from . import app
from . import db


def retryable_transaction(attempts=3, exceptions=(OperationalError,)):
    """Decorator retries a function when expected exceptions are raised."""
    assert len(exceptions) > 0
    assert attempts > 0

    def wrapper(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            for i in xrange(attempts):
                try:
                    return f(*args, **kwargs)
                except exceptions, e:
                    if i == (attempts - 1):
                        raise
                    logging.warning(
                        'Retryable error in transaction on attempt %d. %s: %s',
                        i + 1, e.__class__.__name__, e)
                    db.session.rollback()

        return wrapped

    return wrapper


def jsonify_assert(asserted, message, status_code=400):
    """Asserts something is true, aborts the request if not."""
    if asserted:
        return
    try:
        raise AssertionError(message)
    except AssertionError, e:
        stack = traceback.extract_stack()
        stack.pop()
        logging.error('Assertion failed: %s\n%s',
                      str(e), ''.join(traceback.format_list(stack)))
        abort(jsonify_error(e, status_code=status_code))


def jsonify_error(message_or_exception, status_code=400):
    """Returns a JSON payload that indicates the request had an error."""
    if isinstance(message_or_exception, Exception):
        message = '%s: %s' % (
            message_or_exception.__class__.__name__, message_or_exception)
    else:
        message = message_or_exception

    logging.debug('Returning status=%d, error message: %s',
                  status_code, message)
    response = jsonify(error=message)
    response.status_code = status_code
    return response


def ignore_exceptions(f):
    """Decorator catches and ignores any exceptions raised by this function."""
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except:
            logging.exception("Ignoring exception in %r", f)
    return wrapped


# Based on http://flask.pocoo.org/snippets/33/
@app.template_filter()
def timesince(when):
    """Returns string representing "time since" or "time until".

    Examples:
        3 days ago, 5 hours ago, 3 minutes from now, 5 hours from now, now.
    """
    if not when:
        return ''

    now = datetime.datetime.utcnow()
    if now > when:
        diff = now - when
        suffix = 'ago'
    else:
        diff = when - now
        suffix = 'from now'

    periods = (
        (diff.days / 365, 'year', 'years'),
        (diff.days / 30, 'month', 'months'),
        (diff.days / 7, 'week', 'weeks'),
        (diff.days, 'day', 'days'),
        (diff.seconds / 3600, 'hour', 'hours'),
        (diff.seconds / 60, 'minute', 'minutes'),
        (diff.seconds, 'second', 'seconds'),
    )

    for period, singular, plural in periods:
        if period:
            return '%d %s %s' % (
                period,
                singular if period == 1 else plural,
                suffix)

    return 'now'


def human_uuid():
    """Returns a good UUID for using as a human readable string."""
    return base64.b32encode(
        hashlib.sha1(uuid.uuid4().bytes).digest()).lower().strip('=')


def password_uuid():
    """Returns a good UUID for using as a password."""
    return base64.b64encode(
        hashlib.sha1(uuid.uuid4().bytes).digest()).strip('=')


def is_production():
    """Returns True if this is the production environment."""
    # TODO: Support other deployment situations.
    return (
        os.environ.get('SERVER_SOFTWARE', '').startswith('Google App Engine'))


def get_deployment_timestamp():
    """Returns a unique string represeting the current deployment.

    Used for busting caches.
    """
    # TODO: Support other deployment situations.
    if os.environ.get('SERVER_SOFTWARE', '').startswith('Google App Engine'):
        version_id = os.environ.get('CURRENT_VERSION_ID')
        major_version, timestamp = version_id.split('.', 1)
        return timestamp
    return 'test'


# From http://flask.pocoo.org/snippets/53/
def after_this_request(func):
    if not hasattr(g, 'call_after_request'):
        g.call_after_request = []
    g.call_after_request.append(func)
    return func


def per_request_callbacks(sender, response):
    for func in getattr(g, 'call_after_request', ()):
        response = func(response)
    return response


# Do this with signals instead of @app.after_request because the session
# cookie is added to the response *after* "after_request" callbacks run.
flask.request_finished.connect(per_request_callbacks, app)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Main module for the API server."""

import datetime
import logging
import os

# Local libraries
from flask import Flask, url_for
from flask.ext.cache import Cache
from flask.ext.login import LoginManager
from flask.ext.mail import Mail
from flask.ext.sqlalchemy import SQLAlchemy
import jinja2

# Local modules required for app setup
import config


app = Flask(__name__)
app.config.from_object(config)


db = SQLAlchemy(
    app,
    # Don't expire model instances on commit. Let functions continue to
    # quickly read properties from their last known-good state.
    session_options=dict(expire_on_commit=False))


login = LoginManager(app)
login.login_view = 'login_view'
login.refresh_view = 'login_view'


cache = Cache(app)


mail = Mail(app)


# Modules with handlers to register with the app
from dpxdt.server import api
from dpxdt.server import auth
from dpxdt.server import emails
from dpxdt.server import frontend
from dpxdt.server import work_queue
from dpxdt.server import work_queue_handlers



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Models for managing screenshots and incremental perceptual diffs."""

import datetime

# Local modules
from . import app
from . import db


class User(db.Model):
    """Represents a user who is authenticated in the system.

    Primary key is prefixed with a valid AUTH_TYPES like:

        'google_oauth2:1234567890'
    """

    EMAIL_INVITATION = 'email_invitation'
    GOOGLE_OAUTH2 = 'google_oauth2'
    AUTH_TYPES = frozenset([EMAIL_INVITATION, GOOGLE_OAUTH2])

    id = db.Column(db.String(255), primary_key=True)
    created = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    last_seen = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    email_address = db.Column(db.String(255))
    superuser = db.Column(db.Boolean, default=False)

    def get_auth_type(self):
        return self.id.split(':', 1)[0]

    # For flask-cache memoize key.
    def __repr__(self):
        return 'User(id=%r)' % self.get_id()

    # Methods required by flask-login.
    def is_authenticated(self):
        return True

    def is_active(self):
        return True

    def is_anonymous(self):
        return False

    def get_id(self):
        return self.id

    def __eq__(self, other):
        return other.id == self.id

    def __ne__(self, other):
        return other.id != self.id


class ApiKey(db.Model):
    """API access for an automated system."""

    id = db.Column(db.String(255), primary_key=True)
    secret = db.Column(db.String(255), nullable=False)
    active = db.Column(db.Boolean, default=True)
    purpose = db.Column(db.String(255))
    created = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    modified = db.Column(db.DateTime, default=datetime.datetime.utcnow,
                         onupdate=datetime.datetime.utcnow)
    revoked = db.Column(db.DateTime)
    superuser = db.Column(db.Boolean, default=False)
    build_id = db.Column(db.Integer, db.ForeignKey('build.id'))


ownership_table = db.Table(
    'build_ownership',
    db.Column('build_id', db.Integer, db.ForeignKey('build.id')),
    db.Column('user_id', db.String(255), db.ForeignKey('user.id')))


class Build(db.Model):
    """A single repository of artifacts and diffs owned by someone."""

    id = db.Column(db.Integer, primary_key=True)
    created = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    modified = db.Column(db.DateTime, default=datetime.datetime.utcnow,
                         onupdate=datetime.datetime.utcnow)
    name = db.Column(db.String(255))
    public = db.Column(db.Boolean, default=False)
    owners = db.relationship('User', secondary=ownership_table,
                             backref=db.backref('builds', lazy='dynamic'),
                             lazy='dynamic')
    send_email = db.Column(db.Boolean, default=True)
    email_alias = db.Column(db.String(255))

    def is_owned_by(self, user_id):
        return self.owners.filter_by(id=user_id).first() is not None

    # For flask-cache memoize key.
    def __repr__(self):
        return 'Build(id=%r)' % self.id


class Release(db.Model):
    """A set of runs in a build, grouped by user-supplied name."""

    RECEIVING = 'receiving'
    PROCESSING = 'processing'
    REVIEWING = 'reviewing'
    BAD = 'bad'
    GOOD = 'good'
    STATES = frozenset([RECEIVING, PROCESSING, REVIEWING, BAD, GOOD])

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), nullable=False)
    number = db.Column(db.Integer, nullable=False)
    created = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    modified = db.Column(db.DateTime, default=datetime.datetime.utcnow,
                         onupdate=datetime.datetime.utcnow)
    status = db.Column(db.Enum(*STATES), default=RECEIVING, nullable=False)
    build_id = db.Column(db.Integer, db.ForeignKey('build.id'), nullable=False)
    url = db.Column(db.String(2048))

    # For flask-cache memoize key.
    def __repr__(self):
        return 'Release(id=%r)' % self.id


artifact_ownership_table = db.Table(
    'artifact_ownership',
    db.Column('artifact', db.String(100), db.ForeignKey('artifact.id')),
    db.Column('build_id', db.Integer, db.ForeignKey('build.id')))


class Artifact(db.Model):
    """Contains a single file uploaded by a diff worker."""

    id = db.Column(db.String(100), primary_key=True)
    created = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    data = db.Column(db.LargeBinary(length=2**31))
    alternate = db.Column(db.Text)
    content_type = db.Column(db.String(255))
    owners = db.relationship('Build', secondary=artifact_ownership_table,
                             backref=db.backref('artifacts', lazy='dynamic'),
                             lazy='dynamic')


class Run(db.Model):
    """Contains a set of screenshot records uploaded by a diff worker."""

    DATA_PENDING = 'data_pending'
    DIFF_APPROVED = 'diff_approved'
    DIFF_FOUND = 'diff_found'
    DIFF_NOT_FOUND = 'diff_not_found'
    FAILED = 'failed'
    NEEDS_DIFF = 'needs_diff'
    NO_DIFF_NEEDED = 'no_diff_needed'

    STATES = frozenset([
        DATA_PENDING, DIFF_APPROVED, DIFF_FOUND, DIFF_NOT_FOUND,
        FAILED, NEEDS_DIFF, NO_DIFF_NEEDED])

    DIFF_NEEDED_STATES = frozenset([DIFF_FOUND, DIFF_APPROVED])

    id = db.Column(db.Integer, primary_key=True)
    release_id = db.Column(db.Integer, db.ForeignKey('release.id'))
    release = db.relationship('Release',
                              backref=db.backref('runs', lazy='select'),
                              lazy='joined',
                              join_depth=1)

    name = db.Column(db.String(255), nullable=False)
    # TODO: Put rigid DB constraint on uniqueness of (release_id, name)
    created = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    modified = db.Column(db.DateTime, default=datetime.datetime.utcnow,
                         onupdate=datetime.datetime.utcnow)
    status = db.Column(db.Enum(*STATES), nullable=False)

    image = db.Column(db.String(100), db.ForeignKey('artifact.id'))
    log = db.Column(db.String(100), db.ForeignKey('artifact.id'))
    config = db.Column(db.String(100), db.ForeignKey('artifact.id'))
    url = db.Column(db.String(2048))

    ref_image = db.Column(db.String(100), db.ForeignKey('artifact.id'))
    ref_log = db.Column(db.String(100), db.ForeignKey('artifact.id'))
    ref_config = db.Column(db.String(100), db.ForeignKey('artifact.id'))
    ref_url = db.Column(db.String(2048))

    diff_image = db.Column(db.String(100), db.ForeignKey('artifact.id'))
    diff_log = db.Column(db.String(100), db.ForeignKey('artifact.id'))
    distortion = db.Column(db.Float())

    tasks = db.relationship('WorkQueue',
                            backref=db.backref('runs', lazy='select'),
                            lazy='joined',
                            join_depth=1,
                            order_by='WorkQueue.created')

    # For flask-cache memoize key.
    def __repr__(self):
        return 'Run(id=%r)' % self.id


class AdminLog(db.Model):
    """Log of admin user actions for a build."""

    CHANGED_SETTINGS = 'changed_settings'
    CREATED_API_KEY = 'created_api_key'
    CREATED_BUILD = 'created_build'
    INVITE_ACCEPTED = 'invite_accepted'
    INVITED_NEW_ADMIN = 'invited_new_admin'
    REVOKED_ADMIN = 'revoked_admin'
    REVOKED_API_KEY = 'revoked_api_key'
    RUN_APPROVED = 'run_approved'
    RUN_REJECTED = 'run_rejected'
    RELEASE_BAD = 'release_bad'
    RELEASE_GOOD = 'release_good'
    RELEASE_REVIEWING = 'release_reviewing'

    LOG_TYPES = frozenset([
        CHANGED_SETTINGS, CREATED_API_KEY, CREATED_BUILD, INVITE_ACCEPTED,
        INVITED_NEW_ADMIN, REVOKED_ADMIN, REVOKED_API_KEY, RUN_APPROVED,
        RUN_REJECTED, RELEASE_BAD, RELEASE_GOOD, RELEASE_REVIEWING])

    id = db.Column(db.Integer, primary_key=True)
    build_id = db.Column(db.Integer, db.ForeignKey('build.id'), nullable=False)

    release_id = db.Column(db.Integer, db.ForeignKey('release.id'))
    release = db.relationship('Release', lazy='joined', join_depth=2)

    run_id = db.Column(db.Integer, db.ForeignKey('run.id'))
    run = db.relationship('Run', lazy='joined', join_depth=1)

    user_id = db.Column(db.String(255), db.ForeignKey('user.id'))
    user = db.relationship('User', lazy='joined', join_depth=1)

    created = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    log_type = db.Column(db.Enum(*LOG_TYPES), nullable=False)
    message = db.Column(db.Text)

    # For flask-cache memoize key.
    def __repr__(self):
        return 'AdminLog(id=%r)' % self.id



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Workers for driving screen captures, perceptual diffs, and related work."""

import Queue
import base64
import json
import logging
import shutil
import socket
import ssl
import time
import urllib
import urllib2

# Local Libraries
import gflags
FLAGS = gflags.FLAGS
import poster.encode
import poster.streaminghttp
poster.streaminghttp.register_openers()

# Local modules
from dpxdt.client import workers


gflags.DEFINE_float(
    'fetch_frequency', 1.0,
    'Maximum number of fetches to make per second per thread.')

gflags.DEFINE_integer(
    'fetch_threads', 1, 'Number of fetch threads to run')


class FetchItem(workers.WorkItem):
    """Work item that is handled by fetching a URL."""

    def __init__(self, url, post=None, timeout_seconds=30, result_path=None,
                 username=None, password=None):
        """Initializer.

        Args:
            url: URL to fetch.
            post: Optional. Dictionary of post parameters to include in the
                request, with keys and values coerced to strings. If any
                values are open file handles, the post data will be formatted
                as multipart/form-data.
            timeout_seconds: Optional. How long until the fetch should timeout.
            result_path: When supplied, the output of the fetch should be
                streamed to a file on disk with the given path. Use this
                to prevent many fetches from causing memory problems.
            username: Optional. Username to use for the request, for
                HTTP basic authentication.
            password: Optional. Password to use for the request, for
                HTTP basic authentication.
        """
        workers.WorkItem.__init__(self)
        self.url = url
        self.post = post
        self.username = username
        self.password = password
        self.timeout_seconds = timeout_seconds
        self.result_path = result_path
        self.status_code = None
        self.data = None
        self.headers = None
        self._data_json = None

    def _get_dict_for_repr(self):
        result = self.__dict__.copy()
        if result.get('password'):
            result['password'] = 'ELIDED'
        return result

    @property
    def json(self):
        """Returns de-JSONed data or None if it's a different content type."""
        if self._data_json:
            return self._data_json

        if not self.data or self.headers.gettype() != 'application/json':
            return None

        self._data_json = json.loads(self.data)
        return self._data_json


class FetchThread(workers.WorkerThread):
    """Worker thread for fetching URLs."""

    def handle_item(self, item):
        start_time = time.time()

        if item.post is not None:
            adjusted_data = {}
            use_form_data = False

            for key, value in item.post.iteritems():
                if value is None:
                    continue
                if isinstance(value, file):
                    use_form_data = True
                adjusted_data[key] = value

            if use_form_data:
                datagen, headers = poster.encode.multipart_encode(
                    adjusted_data)
                request = urllib2.Request(item.url, datagen, headers)
            else:
                request = urllib2.Request(
                    item.url, urllib.urlencode(adjusted_data))
        else:
            request = urllib2.Request(item.url)

        if item.username:
            credentials = base64.b64encode(
                '%s:%s' % (item.username, item.password))
            request.add_header('Authorization', 'Basic %s' % credentials)

        try:
            try:
                conn = urllib2.urlopen(request, timeout=item.timeout_seconds)
            except urllib2.HTTPError, e:
                conn = e
            except (urllib2.URLError, ssl.SSLError), e:
                # TODO: Make this status more clear
                item.status_code = 400
                return item

            try:
                item.status_code = conn.getcode()
                item.headers = conn.info()
                if item.result_path:
                    with open(item.result_path, 'wb') as result_file:
                        shutil.copyfileobj(conn, result_file)
                else:
                    item.data = conn.read()
            except socket.timeout, e:
                # TODO: Make this status more clear
                item.status_code = 400
                return item
            finally:
                conn.close()

            return item
        finally:
            end_time = time.time()
            wait_duration = (1.0 / FLAGS.fetch_frequency) - (
                end_time - start_time)
            if wait_duration > 0:
                logging.debug('Rate limiting URL fetch for %f seconds',
                              wait_duration)
                time.sleep(wait_duration)


def register(coordinator):
    """Registers this module as a worker with the given coordinator."""
    fetch_queue = Queue.Queue()
    coordinator.register(FetchItem, fetch_queue)
    for i in xrange(FLAGS.fetch_threads):
        coordinator.worker_threads.append(
            FetchThread(fetch_queue, coordinator.input_queue))



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Background worker that uploads new release candidates."""

import hashlib
import os

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

# Local modules
from dpxdt.client import fetch_worker
from dpxdt.client import workers


gflags.DEFINE_string(
    'release_server_prefix', None,
    'URL prefix of where the release server is located, such as '
    '"https://www.example.com/here/is/my/api". This should use HTTPS if '
    'possible, since API requests send credentials using HTTP basic auth.')

gflags.DEFINE_string(
    'release_client_id', None,
    'Client ID of the API key to use for requests to the release server.')

gflags.DEFINE_string(
    'release_client_secret', None,
    'Client secret of the API key to use for requests to the release server.')


class Error(Exception):
    """Base-class for exceptions in this module."""

class CreateReleaseError(Error):
    """Creating a new release failed for some reason."""

class UploadFileError(Error):
    """Uploading a file failed for some reason."""

class FindRunError(Error):
    """Finding a run failed for some reason."""

class RequestRunError(Error):
    """Requesting a run failed for some reason."""

class ReportRunError(Error):
    """Reporting a run failed for some reason."""

class ReportPdiffError(Error):
    """Reporting a pdiff failed for some reason."""

class RunsDoneError(Error):
    """Marking that all runs are done failed for some reason."""

class DownloadArtifactError(Error):
    """Downloading an artifact failed for some reason."""


class StreamingSha1File(file):
    """File sub-class that sha1 hashes the data as it's read."""

    def __init__(self, *args, **kwargs):
        """Replacement for open()."""
        file.__init__(self, *args, **kwargs)
        self.sha1 = hashlib.sha1()

    def read(self, *args):
        data = file.read(self, *args)
        self.sha1.update(data)
        return data

    def close(self):
        file.close(self)

    def hexdigest(self):
        return self.sha1.hexdigest()


class CreateReleaseWorkflow(workers.WorkflowItem):
    """Creates a new release candidate.

    Args:
        build_id: ID of the build.
        release_name: Name of the release candidate.
        url: Landing URL of the new release.

    Returns:
        The newly created release_number.

    Raises:
        CreateReleaseError if the release could not be created.
    """

    def run(self, build_id, release_name, url):
        call = yield fetch_worker.FetchItem(
            FLAGS.release_server_prefix + '/create_release',
            post={
                'build_id': build_id,
                'release_name': release_name,
                'url': url,
            },
            username=FLAGS.release_client_id,
            password=FLAGS.release_client_secret)

        if call.json and call.json.get('error'):
            raise CreateReleaseError(call.json.get('error'))

        if not call.json or not call.json.get('release_number'):
            raise CreateReleaseError('Bad response: %r' % call)

        raise workers.Return(call.json['release_number'])


class UploadFileWorkflow(workers.WorkflowItem):
    """Uploads a file for a build.

    Args:
        build_id: ID of the build to upload a file for.
        file_path: Path to the file to upload.

    Returns:
        sha1 sum of the file's contents or None if the file could not
        be found.

    Raises:
        UploadFileError if the file could not be uploaded.
    """

    def run(self, build_id, file_path):
        try:
            handle = StreamingSha1File(file_path, 'rb')
            upload = yield fetch_worker.FetchItem(
                FLAGS.release_server_prefix + '/upload',
                post={'build_id': build_id, 'file': handle},
                timeout_seconds=120,
                username=FLAGS.release_client_id,
                password=FLAGS.release_client_secret)

            if upload.json and upload.json.get('error'):
                raise UploadFileError(upload.json.get('error'))

            sha1sum = handle.hexdigest()
            if not upload.json or upload.json.get('sha1sum') != sha1sum:
                raise UploadFileError('Bad response: %r' % upload)

            raise workers.Return(sha1sum)

        except IOError:
            raise workers.Return(None)


class FindRunWorkflow(workers.WorkflowItem):
    """Finds the last good run for a release.

    Args:
        build_id: ID of the build.
        run_name: Name of the run being uploaded.

    Returns:
        JSON dictionary representing the run that was found, with the keys:
        build_id, release_name, release_number, run_name, url, image, log,
        config.

    Raises:
        FindRunError if a run could not be found.
    """

    def run(self, build_id, run_name):
        call = yield fetch_worker.FetchItem(
            FLAGS.release_server_prefix + '/find_run',
            post={
                'build_id': build_id,
                'run_name': run_name,
            },
            username=FLAGS.release_client_id,
            password=FLAGS.release_client_secret)

        if call.json and call.json.get('error'):
            raise FindRunError(call.json.get('error'))

        if not call.json:
            raise FindRunError('Bad response: %r' % call)

        raise workers.Return(call.json)


class RequestRunWorkflow(workers.WorkflowItem):
    """Requests the API server to do a test run and capture the results.

    Args:
        build_id: ID of the build.
        release_name: Name of the release.
        release_number: Number of the release candidate.
        run_name: Name of the run being requested.
        url: URL to fetch for the run.
        config_data: The JSON data that is the config for this run.
        ref_url: Optional. URL of the baseline to fetch for the run.
        ref_config_data: Optional. The JSON data that is the config for the
            baseline of this run.

    Raises:
        RequestRunError if the run could not be requested.
    """

    def run(self, build_id, release_name, release_number, run_name,
            url=None, config_data=None, ref_url=None, ref_config_data=None):
        post = {
            'build_id': build_id,
            'release_name': release_name,
            'release_number': release_number,
            'run_name': run_name,
            'url': url,
            'config': config_data,
        }
        if ref_url and ref_config_data:
            post.update(
                ref_url=ref_url,
                ref_config=ref_config_data)

        call = yield fetch_worker.FetchItem(
            FLAGS.release_server_prefix + '/request_run',
            post=post,
            username=FLAGS.release_client_id,
            password=FLAGS.release_client_secret)

        if call.json and call.json.get('error'):
            raise RequestRunError(call.json.get('error'))

        if not call.json or not call.json.get('success'):
            raise RequestRunError('Bad response: %r' % call)


class ReportRunWorkflow(workers.WorkflowItem):
    """Reports a run as finished.

    Args:
        build_id: ID of the build.
        release_name: Name of the release.
        release_number: Number of the release candidate.
        run_name: Name of the run being uploaded.
        log_path: Optional. Path to the screenshot log to upload.
        image_path: Optional. Path to the screenshot to upload.
        url: Optional. URL that was fetched for the run.
        config_path: Optional. Path to the config to upload.
        ref_url: Optional. Previously fetched URL this is being compared to.
        ref_image: Optional. Asset ID of the image to compare to.
        ref_log: Optional. Asset ID of the reference image's log.
        ref_config: Optional. Asset ID of the reference image's config.
        baseline: Optional. When specified and True, the log_path, url,
            and image_path are for the reference baseline of the specified
            run, not the new capture. If this is True, the ref_* parameters
            must not be provided.
        run_failed: Optional. When specified and True it means that this run
            has failed for some reason. The run may be tried again in the
            future but this will cause this run to immediately show up as
            failing. When not specified or False the run will be assumed to
            have been successful.

    Raises:
        ReportRunError if the run could not be reported.
    """

    def run(self, build_id, release_name, release_number, run_name,
            image_path=None, log_path=None, url=None, config_path=None,
            ref_url=None, ref_image=None, ref_log=None, ref_config=None,
            baseline=None, run_failed=False):
        if baseline and (ref_url or ref_image or ref_log or ref_config):
            raise ReportRunError(
                'Cannot specify "baseline" along with any "ref_*" arguments.')

        upload_jobs = [
            UploadFileWorkflow(build_id, log_path),
        ]
        if image_path:
            image_index = len(upload_jobs)
            upload_jobs.append(UploadFileWorkflow(build_id, image_path))

        if config_path:
            config_index = len(upload_jobs)
            upload_jobs.append(UploadFileWorkflow(build_id, config_path))

        results = yield upload_jobs
        log_id = results[0]
        image_id = None
        config_id = None
        if image_path:
            image_id = results[image_index]
        if config_path:
            config_id = results[config_index]

        post = {
            'build_id': build_id,
            'release_name': release_name,
            'release_number': release_number,
            'run_name': run_name,
        }

        if baseline:
            ref_url = url
            ref_log = log_id
            ref_image = image_id
            ref_config = config_id
            url = None
            log_id = None
            image_id = None
            config_id = None

        if url:
            post.update(url=url)
        if image_id:
            post.update(image=image_id)
        if log_id:
            post.update(log=log_id)
        if config_id:
            post.update(config=config_id)

        if run_failed:
            post.update(run_failed='yes')

        if ref_url:
            post.update(ref_url=ref_url)
        if ref_image:
            post.update(ref_image=ref_image)
        if ref_log:
            post.update(ref_log=ref_log)
        if ref_config:
            post.update(ref_config=ref_config)

        call = yield fetch_worker.FetchItem(
            FLAGS.release_server_prefix + '/report_run',
            post=post,
            username=FLAGS.release_client_id,
            password=FLAGS.release_client_secret)

        if call.json and call.json.get('error'):
            raise ReportRunError(call.json.get('error'))

        if not call.json or not call.json.get('success'):
            raise ReportRunError('Bad response: %r' % call)


class ReportPdiffWorkflow(workers.WorkflowItem):
    """Reports a pdiff's result status.

    Args:
        build_id: ID of the build.
        release_name: Name of the release.
        release_number: Number of the release candidate.
        run_name: Name of the pdiff being uploaded.
        diff_path: Path to the diff to upload.
        log_path: Path to the diff log to upload.
        diff_failed: True when there was a problem computing the diff. False
            when the diff was computed successfully. Defaults to False.

    Raises:
        ReportPdiffError if the pdiff status could not be reported.
    """

    def run(self, build_id, release_name, release_number, run_name,
            diff_path=None, log_path=None, diff_failed=False, distortion=None):
        diff_id = None
        log_id = None
        if (isinstance(diff_path, basestring) and
                os.path.isfile(diff_path) and
                isinstance(log_path, basestring) and
                os.path.isfile(log_path)):
            diff_id, log_id = yield [
                UploadFileWorkflow(build_id, diff_path),
                UploadFileWorkflow(build_id, log_path),
            ]
        elif isinstance(log_path, basestring) and os.path.isfile(log_path):
            log_id = yield UploadFileWorkflow(build_id, log_path)

        post = {
            'build_id': build_id,
            'release_name': release_name,
            'release_number': release_number,
            'run_name': run_name,
        }
        if diff_id:
            post.update(diff_image=diff_id)
        if log_id:
            post.update(diff_log=log_id)
        if diff_failed:
            post.update(diff_failed='yes')
        if distortion:
            post.update(distortion=distortion)

        call = yield fetch_worker.FetchItem(
            FLAGS.release_server_prefix + '/report_run',
            post=post,
            username=FLAGS.release_client_id,
            password=FLAGS.release_client_secret)

        if call.json and call.json.get('error'):
            raise ReportPdiffError(call.json.get('error'))

        if not call.json or not call.json.get('success'):
            raise ReportPdiffError('Bad response: %r' % call)


class RunsDoneWorkflow(workers.WorkflowItem):
    """Reports all runs are done for a release candidate.

    Args:
        build_id: ID of the build.
        release_name: Name of the release.
        release_number: Number of the release candidate.

    Returns:
        URL of where the results for this release candidate can be viewed.

    Raises:
        RunsDoneError if the release candidate could not have its runs
        marked done.
    """

    def run(self, build_id, release_name, release_number):
        call = yield fetch_worker.FetchItem(
            FLAGS.release_server_prefix + '/runs_done',
            post={
                'build_id': build_id,
                'release_name': release_name,
                'release_number': release_number,
            },
            username=FLAGS.release_client_id,
            password=FLAGS.release_client_secret)

        if call.json and call.json.get('error'):
            raise RunsDoneError(call.json.get('error'))

        if not call.json or not call.json.get('success'):
            raise RunsDoneError('Bad response: %r' % call)

        raise workers.Return(call.json['results_url'])


class DownloadArtifactWorkflow(workers.WorkflowItem):
    """Downloads an artifact to a given path.

    Args:
        build_id: ID of the build.
        sha1sum: Content hash of the artifact to fetch.
        result_path: Path where the artifact should be saved on disk.

    Raises:
        DownloadArtifactError if the artifact could not be found or
        fetched for some reason.
    """

    def run(self, build_id, sha1sum, result_path):
        download_url = '%s/download?sha1sum=%s&build_id=%s' % (
            FLAGS.release_server_prefix, sha1sum, build_id)
        call = yield fetch_worker.FetchItem(
            download_url,
            result_path=result_path,
            username=FLAGS.release_client_id,
            password=FLAGS.release_client_secret)
        if call.status_code != 200:
            raise DownloadArtifactError('Bad response: %r' % call)



#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Runs the dpxdt API server, optionally with local queue workers."""

import logging
import os
import sys
import threading

# Local Libraries
import gflags
FLAGS = gflags.FLAGS

# Local modules
from dpxdt import runworker
from dpxdt import server


gflags.DEFINE_bool(
    'local_queue_workers', False,
    'When true, run queue worker threads locally in the same process '
    'as the server.')

gflags.DEFINE_bool(
    'reload_code', False,
    'Reload code on every request. Should only be used in local development.')

gflags.DEFINE_bool(
    'ignore_auth', False,
    'Ignore any need for authentication for API and frontend accesses. You '
    'should only do this for local development!')

gflags.DEFINE_integer('port', 5000, 'Port to run the HTTP server on.')

gflags.DEFINE_string('host', '0.0.0.0', 'Host argument for the server.')


def main(argv):
    try:
        argv = FLAGS(argv)
    except gflags.FlagsError, e:
        print '%s\nUsage: %s ARGS\n%s' % (e, sys.argv[0], FLAGS)
        sys.exit(1)

    logging.basicConfig(
        format='%(levelname)s %(filename)s:%(lineno)s] %(message)s')
    if FLAGS.verbose:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)

    if FLAGS.verbose_queries:
        logging.getLogger('sqlalchemy.engine').setLevel(logging.DEBUG)

    if FLAGS.local_queue_workers:
        coordinator = runworker.run_workers()

        # If the babysitter thread dies, the whole process goes down.
        def worker_babysitter():
            try:
                coordinator.wait_one()
            finally:
                os._exit(1)

        babysitter_thread = threading.Thread(target=worker_babysitter)
        babysitter_thread.setDaemon(True)
        babysitter_thread.start()

    if FLAGS.ignore_auth:
        server.app.config['IGNORE_AUTH'] = True

    server.app.run(debug=FLAGS.reload_code, host=FLAGS.host, port=FLAGS.port)


if __name__ == '__main__':
    main(sys.argv)
